"""
Service support for Debian systems (uses update-rc.d and /sbin/service)

.. important::
    If you feel that Salt should be using this module to manage services on a
    minion, and it is using a different module (or gives an error similar to
    *'service.start' is not available*), see :ref:`here
    <module-provider-override>`.
"""

import fnmatch
import glob
import logging
import os
import re
import shlex

import salt.utils.systemd

__func_alias__ = {"reload_": "reload"}

# Define the module's virtual name
__virtualname__ = "service"

log = logging.getLogger(__name__)


def __virtual__():
    """
    Only work on Debian and when systemd isn't running
    """
    if __grains__["os"] in (
        "Debian",
        "Raspbian",
        "Devuan",
        "NILinuxRT",
    ) and not salt.utils.systemd.booted(__context__):
        return __virtualname__
    else:
        return (
            False,
            "The debian_service module could not be loaded: "
            "unsupported OS family and/or systemd running.",
        )


def _service_cmd(*args):
    return "service {} {}".format(args[0], " ".join(args[1:]))


def _get_runlevel():
    """
    returns the current runlevel
    """
    out = __salt__["cmd.run"]("runlevel")
    # unknown can be returned while inside a container environment, since
    # this is due to a lack of init, it should be safe to assume runlevel
    # 2, which is Debian's default. If not, all service related states
    # will throw an out of range exception here which will cause
    # other functions to fail.
    if "unknown" in out:
        return "2"
    else:
        return out.split()[1]


def get_enabled():
    """
    Return a list of service that are enabled on boot

    CLI Example:

    .. code-block:: bash

        salt '*' service.get_enabled
    """
    prefix = f"/etc/rc[S{_get_runlevel()}].d/S"
    ret = set()
    for line in [x.rsplit(os.sep, 1)[-1] for x in glob.glob(f"{prefix}*")]:
        ret.add(re.split(r"\d+", line)[-1])
    return sorted(ret)


def get_disabled():
    """
    Return a set of services that are installed but disabled

    CLI Example:

    .. code-block:: bash

        salt '*' service.get_disabled
    """
    return sorted(set(get_all()) - set(get_enabled()))


def available(name):
    """
    Returns ``True`` if the specified service is available, otherwise returns
    ``False``.

    CLI Example:

    .. code-block:: bash

        salt '*' service.available sshd
    """
    return name in get_all()


def missing(name):
    """
    The inverse of service.available.
    Returns ``True`` if the specified service is not available, otherwise returns
    ``False``.

    CLI Example:

    .. code-block:: bash

        salt '*' service.missing sshd
    """
    return name not in get_all()


def get_all():
    """
    Return all available boot services

    CLI Example:

    .. code-block:: bash

        salt '*' service.get_all
    """
    ret = set()
    lines = glob.glob("/etc/init.d/*")
    for line in lines:
        service = line.split("/etc/init.d/")[1]
        # Remove README.  If it's an enabled service, it will be added back in.
        if service != "README":
            ret.add(service)
    return sorted(ret | set(get_enabled()))


def start(name):
    """
    Start the specified service

    CLI Example:

    .. code-block:: bash

        salt '*' service.start <service name>
    """
    cmd = _service_cmd(name, "start")
    return not __salt__["cmd.retcode"](cmd)


def stop(name):
    """
    Stop the specified service

    CLI Example:

    .. code-block:: bash

        salt '*' service.stop <service name>
    """
    cmd = _service_cmd(name, "stop")
    return not __salt__["cmd.retcode"](cmd)


def restart(name):
    """
    Restart the named service

    CLI Example:

    .. code-block:: bash

        salt '*' service.restart <service name>
    """
    cmd = _service_cmd(name, "restart")
    return not __salt__["cmd.retcode"](cmd)


def reload_(name):
    """
    Reload the named service

    CLI Example:

    .. code-block:: bash

        salt '*' service.reload <service name>
    """
    cmd = _service_cmd(name, "reload")
    return not __salt__["cmd.retcode"](cmd)


def force_reload(name):
    """
    Force-reload the named service

    CLI Example:

    .. code-block:: bash

        salt '*' service.force_reload <service name>
    """
    cmd = _service_cmd(name, "force-reload")
    return not __salt__["cmd.retcode"](cmd)


def status(name, sig=None):
    """
    Return the status for a service.
    If the name contains globbing, a dict mapping service name to True/False
    values is returned.

    .. versionchanged:: 2018.3.0
        The service name can now be a glob (e.g. ``salt*``)

    Args:
        name (str): The name of the service to check
        sig (str): Signature to use to find the service via ps

    Returns:
        bool: True if running, False otherwise
        dict: Maps service name to True if running, False otherwise

    CLI Example:

    .. code-block:: bash

        salt '*' service.status <service name> [service signature]
    """
    if sig:
        return bool(__salt__["status.pid"](sig))

    contains_globbing = bool(re.search(r"\*|\?|\[.+\]", name))
    if contains_globbing:
        services = fnmatch.filter(get_all(), name)
    else:
        services = [name]
    results = {}
    for service in services:
        cmd = _service_cmd(service, "status")
        results[service] = not __salt__["cmd.retcode"](cmd, ignore_retcode=True)
    if contains_globbing:
        return results
    return results[name]


def enable(name, **kwargs):
    """
    Enable the named service to start at boot

    CLI Example:

    .. code-block:: bash

        salt '*' service.enable <service name>
    """
    cmd = "insserv {0} && update-rc.d {0} enable".format(shlex.quote(name))
    return not __salt__["cmd.retcode"](cmd, python_shell=True)


def disable(name, **kwargs):
    """
    Disable the named service to start at boot

    CLI Example:

    .. code-block:: bash

        salt '*' service.disable <service name>
    """
    cmd = f"update-rc.d {name} disable"
    return not __salt__["cmd.retcode"](cmd)


def enabled(name, **kwargs):
    """
    Return True if the named service is enabled, false otherwise

    CLI Example:

    .. code-block:: bash

        salt '*' service.enabled <service name>
    """
    return name in get_enabled()


def disabled(name):
    """
    Return True if the named service is disabled, false otherwise

    CLI Example:

    .. code-block:: bash

        salt '*' service.disabled <service name>
    """
    return name in get_disabled()



"""
Clone a remote SVN repository and use the filesystem as a Pillar source

This external Pillar source can be configured in the master config file like
so:

.. code-block:: yaml

    ext_pillar:
      - svn: trunk svn://svnserver/repo root=subdirectory

The `root=` parameter is optional and used to set the subdirectory from where
to look for Pillar files (such as ``top.sls``).

.. versionchanged:: 2014.7.0
    The optional ``root`` parameter will be added.

Note that this is not the same thing as configuring pillar data using the
:conf_master:`pillar_roots` parameter. The branch referenced in the
:conf_master:`ext_pillar` entry above (``master``), would evaluate to the
``base`` environment, so this branch needs to contain a ``top.sls`` with a
``base`` section in it, like this:

.. code-block:: yaml

    base:
      '*':
        - foo

To use other environments from the same SVN repo as svn_pillar sources, just
add additional lines, like so:

.. code-block:: yaml

    ext_pillar:
      - svn: trunk svn://svnserver/repo
      - svn: dev svn://svnserver/repo

In this case, the ``dev`` branch would need its own ``top.sls`` with a ``dev``
section in it, like this:

.. code-block:: yaml

    dev:
      '*':
        - bar
"""

import hashlib
import logging
import os
from copy import deepcopy

from salt.pillar import Pillar

HAS_SVN = False
try:
    import pysvn

    HAS_SVN = True
    CLIENT = pysvn.Client()
except ImportError:
    pass


# Set up logging
log = logging.getLogger(__name__)

# Define the module's virtual name
__virtualname__ = "svn"


def __virtual__():
    ext_pillar_sources = [x for x in __opts__.get("ext_pillar", [])]
    if not any(["svn" in x for x in ext_pillar_sources]):
        return False
    if not HAS_SVN:
        log.error(
            "SVN-based ext_pillar is enabled in configuration but "
            "could not be loaded, is pysvn installed?"
        )
        return False
    return __virtualname__


class SvnPillar:
    """
    Deal with the remote SVN repository for Pillar
    """

    def __init__(self, branch, repo_location, root, opts):
        """
        Try to initialize the SVN repo object
        """
        repo_hash = hashlib.md5(repo_location).hexdigest()
        repo_dir = os.path.join(opts["cachedir"], "pillar_svnfs", repo_hash)

        self.branch = branch
        self.root = root
        self.repo_dir = repo_dir
        self.repo_location = repo_location

        if not os.path.isdir(repo_dir):
            os.makedirs(repo_dir)
            log.debug("Checking out fileserver for svn_pillar module")
            try:
                CLIENT.checkout(repo_location, repo_dir)
            except pysvn.ClientError:
                log.error(
                    "Failed to initialize svn_pillar %s %s", repo_location, repo_dir
                )

    def update(self):
        try:
            log.debug("Updating fileserver for svn_pillar module")
            CLIENT.update(self.repo_dir)
        except pysvn.ClientError as exc:
            log.error(
                "Unable to fetch the latest changes from remote %s: %s",
                self.repo_location,
                exc,
            )

    def pillar_dir(self):
        """
        Returns the directory of the pillars (repo cache + branch + root)
        """
        repo_dir = self.repo_dir
        root = self.root
        branch = self.branch
        if branch == "trunk" or branch == "base":
            working_dir = os.path.join(repo_dir, "trunk", root)
            if not os.path.isdir(working_dir):
                log.error("Could not find %s/trunk/%s", self.repo_location, root)
            else:
                return os.path.normpath(working_dir)
        working_dir = os.path.join(repo_dir, "branches", branch, root)
        if os.path.isdir(working_dir):
            return os.path.normpath(working_dir)
        working_dir = os.path.join(working_dir, "tags", branch, root)
        if os.path.isdir(working_dir):
            return os.path.normpath(working_dir)
        log.error("Could not find %s/branches/%s/%s", self.repo_location, branch, root)
        return repo_dir


def _extract_key_val(kv, delimiter="="):
    """Extract key and value from key=val string.

    Example:
    >>> _extract_key_val('foo=bar')
    ('foo', 'bar')
    """
    pieces = kv.split(delimiter)
    key = pieces[0]
    val = delimiter.join(pieces[1:])
    return key, val


def ext_pillar(minion_id, pillar, repo_string):  # pylint: disable=W0613
    """
    Execute a command and read the output as YAML
    """
    # split the branch, repo name and optional extra (key=val) parameters.
    options = repo_string.strip().split()
    branch = options[0]
    repo_location = options[1]
    root = ""

    for extraopt in options[2:]:
        # Support multiple key=val attributes as custom parameters.
        DELIM = "="
        if DELIM not in extraopt:
            log.error(
                "Incorrectly formatted extra parameter. Missing '%s': %s",
                DELIM,
                extraopt,
            )
        key, val = _extract_key_val(extraopt, DELIM)
        if key == "root":
            root = val
        else:
            log.warning("Unrecognized extra parameter: %s", key)

    svnpil = SvnPillar(branch, repo_location, root, __opts__)

    # environment is "different" from the branch
    branch = branch == "trunk" and "base" or branch

    pillar_dir = svnpil.pillar_dir()
    log.debug("[pillar_roots][%s] = %s", branch, pillar_dir)

    # Don't recurse forever-- the Pillar object will re-call the ext_pillar
    # function
    if __opts__["pillar_roots"].get(branch, []) == [pillar_dir]:
        return {}
    svnpil.update()
    opts = deepcopy(__opts__)
    opts["pillar_roots"][branch] = [pillar_dir]
    pil = Pillar(opts, __grains__, minion_id, branch)
    return pil.compile_pillar(ext=False)



import copy
import logging

import pytest
from requests.exceptions import HTTPError

import salt.pillar.vault as vault
from tests.support.mock import Mock, patch


@pytest.fixture
def configure_loader_modules():
    return {
        vault: {
            "__utils__": {
                "vault.expand_pattern_lists": Mock(
                    side_effect=lambda x, *args, **kwargs: [x]
                )
            }
        }
    }


@pytest.fixture
def vault_kvv1():
    res = Mock(status_code=200)
    res.json.return_value = {"data": {"foo": "bar"}}
    return Mock(return_value=res)


@pytest.fixture
def vault_kvv2():
    res = Mock(status_code=200)
    res.json.return_value = {"data": {"data": {"foo": "bar"}}, "metadata": {}}
    return Mock(return_value=res)


@pytest.fixture
def is_v2_false():
    path = "secret/path"
    return {"v2": False, "data": path, "metadata": path, "delete": path, "type": "kv"}


@pytest.fixture
def is_v2_true():
    return {
        "v2": True,
        "data": "secret/data/path",
        "metadata": "secret/metadata/path",
        "type": "kv",
    }


@pytest.mark.parametrize(
    "is_v2,vaultkv", [("is_v2_false", "vault_kvv1"), ("is_v2_true", "vault_kvv2")]
)
def test_ext_pillar(is_v2, vaultkv, request):
    """
    Test ext_pillar functionality for KV v1/2
    """
    is_v2 = request.getfixturevalue(is_v2)
    vaultkv = request.getfixturevalue(vaultkv)
    with patch.dict(
        vault.__utils__,
        {"vault.is_v2": Mock(return_value=is_v2), "vault.make_request": vaultkv},
    ):
        ext_pillar = vault.ext_pillar("testminion", {}, "path=secret/path")
        vaultkv.assert_called_once_with("GET", "v1/" + is_v2["data"])
        assert "foo" in ext_pillar
        assert "metadata" not in ext_pillar
        assert "data" not in ext_pillar
        assert ext_pillar["foo"] == "bar"


def test_ext_pillar_not_found(is_v2_false, caplog):
    """
    Test that HTTP 404 is handled correctly
    """
    res = Mock(status_code=404, ok=False)
    res.raise_for_status.side_effect = HTTPError()
    with caplog.at_level(logging.INFO):
        with patch.dict(
            vault.__utils__,
            {
                "vault.is_v2": Mock(return_value=is_v2_false),
                "vault.make_request": Mock(return_value=res),
            },
        ):
            ext_pillar = vault.ext_pillar("testminion", {}, "path=secret/path")
            assert ext_pillar == {}
            assert "Vault secret not found for: secret/path" in caplog.messages


def test_ext_pillar_nesting_key(is_v2_false, vault_kvv1):
    """
    Test that nesting_key is honored as expected
    """
    with patch.dict(
        vault.__utils__,
        {
            "vault.is_v2": Mock(return_value=is_v2_false),
            "vault.make_request": vault_kvv1,
        },
    ):
        ext_pillar = vault.ext_pillar(
            "testminion", {}, "path=secret/path", nesting_key="baz"
        )
        assert "foo" not in ext_pillar
        assert "baz" in ext_pillar
        assert "foo" in ext_pillar["baz"]
        assert ext_pillar["baz"]["foo"] == "bar"


@pytest.mark.parametrize(
    "pattern,expected",
    [
        ("no/template/in/use", ["no/template/in/use"]),
        ("salt/minions/{minion}", ["salt/minions/test-minion"]),
        ("salt/roles/{pillar[role]}", ["salt/roles/foo"]),
        ("salt/roles/{pillar[nonexistent]}", []),
    ],
)
def test_get_paths(pattern, expected):
    """
    Test that templated paths are resolved as expected.
    Expansion of lists is tested in the utility module unit test.
    """
    previous_pillar = {
        "role": "foo",
    }
    result = vault._get_paths(pattern, "test-minion", previous_pillar)
    assert result == expected


def test_ext_pillar_merging(is_v2_false):
    """
    Test that patterns that result in multiple paths are merged as expected.
    """

    def make_request(method, resource, *args, **kwargs):
        vault_data = {
            "v1/salt/roles/db": {
                "from_db": True,
                "pass": "hunter2",
                "list": ["a", "b"],
            },
            "v1/salt/roles/web": {
                "from_web": True,
                "pass": "hunter1",
                "list": ["c", "d"],
            },
        }
        res = Mock(status_code=200, ok=True)
        res.json.return_value = {"data": copy.deepcopy(vault_data[resource])}
        return res

    cases = [
        (
            ["salt/roles/db", "salt/roles/web"],
            {"from_db": True, "from_web": True, "list": ["c", "d"], "pass": "hunter1"},
        ),
        (
            ["salt/roles/web", "salt/roles/db"],
            {"from_db": True, "from_web": True, "list": ["a", "b"], "pass": "hunter2"},
        ),
    ]
    vaultkv = Mock(side_effect=make_request)

    for expanded_patterns, expected in cases:
        with patch.dict(
            vault.__utils__,
            {
                "vault.make_request": vaultkv,
                "vault.expand_pattern_lists": Mock(return_value=expanded_patterns),
                "vault.is_v2": Mock(return_value=is_v2_false),
            },
        ):
            ext_pillar = vault.ext_pillar(
                "test-minion",
                {"roles": ["db", "web"]},
                conf="path=salt/roles/{pillar[roles]}",
                merge_strategy="smart",
                merge_lists=False,
            )
            assert ext_pillar == expected


def test_ext_pillar_disabled_during_policy_pillar_rendering():
    """
    Ensure ext_pillar returns an empty dict when called during pillar
    template rendering to prevent a cyclic dependency.
    """
    mock_version = Mock()
    mock_vault = Mock()
    extra = {"_vault_runner_is_compiling_pillar_templates": True}

    with patch.dict(
        vault.__utils__, {"vault.make_request": mock_vault, "vault.is_v2": mock_version}
    ):
        assert {} == vault.ext_pillar(
            "test-minion", {}, conf="path=secret/path", extra_minion_data=extra
        )
        assert mock_version.call_count == 0
        assert mock_vault.call_count == 0


def test_invalid_config(caplog):
    """
    Ensure an empty dict is returned and an error is logged in case
    the config does not contain path=<...>
    """
    with caplog.at_level(logging.ERROR):
        ext_pillar = vault.ext_pillar("testminion", {}, "secret/path")
        assert ext_pillar == {}
        assert "is not a valid Vault ext_pillar config" in caplog.text



#!/usr/bin/python
#
# committer_parser.py
#
# Simple script to parse the output of 'git log' and generate some statistics.
# May leverage GitHub API in the future
#
"""
To use this commit parser script pipe git log into the stdin:

    git log | committer_parser.py -c -
"""
# pylint: disable=resource-leakage


import datetime
import email.utils
import getopt
import re
import sys


class Usage(Exception):
    def __init__(self, msg):  # pylint: disable=W0231
        self.msg = "committer_parser.py [-c | --contributor-detail] - | <logfilename>\n"
        self.msg += (
            "   : Parse commit log from git and print number of "
            "commits and unique committers\n"
        )
        self.msg += "   : by month.  Accepts a filename or reads from stdin.\n"
        self.msg += (
            "   : -c | --contributor-detail generates output by "
            "contributor, by month, in a tab-separated table\n"
        )
        if msg:
            self.msg += "\n"
            self.msg += msg


def parse_date(datestr):
    d = email.utils.parsedate(datestr)
    return datetime.datetime(d[0], d[1], d[2], d[3], d[4], d[5], d[6])


def parse_gitlog(filename=None):
    """
    Parse out the gitlog cli data
    """
    results = {}
    commits = {}
    commits_by_contributor = {}

    if not filename or filename == "-":
        fh = sys.stdin
    else:
        fh = open(filename, "r+", encoding="utf-8")

    try:
        commitcount = 0
        for line in fh.readlines():
            line = line.rstrip()
            if line.startswith("commit "):
                new_commit = True
                commitcount += 1
                continue

            if line.startswith("Author:"):
                author = re.match(r"Author:\s+(.*)\s+<(.*)>", line)
                if author:
                    email = author.group(2)
                continue

            if line.startswith("Date:"):

                isodate = re.match(r"Date:\s+(.*)", line)
                d = parse_date(isodate.group(1))
                continue

            if len(line) < 2 and new_commit:
                new_commit = False
                key = f"{d.year}-{str(d.month).zfill(2)}"

                if key not in results:
                    results[key] = []

                if key not in commits:
                    commits[key] = 0

                if email not in commits_by_contributor:
                    commits_by_contributor[email] = {}

                if key not in commits_by_contributor[email]:
                    commits_by_contributor[email][key] = 1
                else:
                    commits_by_contributor[email][key] += 1

                    if email not in results[key]:
                        results[key].append(email)

                    commits[key] += commitcount
                    commitcount = 0
    finally:
        fh.close()
    return (results, commits, commits_by_contributor)


def counts_by_contributor(commits_by_contributor, results):
    output = ""
    dates = sorted(results.keys())
    for d in dates:
        output += f"\t{d}"

    output += "\n"

    for email in sorted(commits_by_contributor.keys()):
        output += f"'{email}"
        for d in dates:
            if d in commits_by_contributor[email]:
                output += f"\t{commits_by_contributor[email][d]}"
            else:
                output += "\t"
        output += "\n"
    return output


def count_results(results, commits):
    result_str = ""
    print("Date\tContributors\tCommits")
    for k in sorted(results.keys()):
        result_str += f"{k}\t{len(results[k])}\t{commits[k]}"
        result_str += "\n"
    return result_str


def main(argv=None):
    if argv is None:
        argv = sys.argv
    try:
        try:
            opts, args = getopt.getopt(argv[1:], "hc", ["help", "contributor-detail"])
            if len(args) < 1:
                raise Usage(
                    "committer_parser.py needs a filename or '-' to read from stdin"
                )
        except getopt.error as msg:
            raise Usage(msg)
    except Usage as err:
        print(err.msg, file=sys.stderr)
        return 2

    if len(opts) > 0:
        if "-h" in opts[0] or "--help" in opts[0]:
            return 0

    data, counts, commits_by_contributor = parse_gitlog(filename=args[0])

    if len(opts) > 0:
        if "-c" or "--contributor-detail":
            print(counts_by_contributor(commits_by_contributor, data))
    else:
        print(count_results(data, counts))


if __name__ == "__main__":
    sys.exit(main())



import os

import pytest

import salt.modules.file as filemod
import salt.states.file as file
from tests.support.mock import MagicMock, call, create_autospec, patch


@pytest.fixture
def configure_loader_modules(minion_opts):
    return {
        file: {
            "__opts__": {"test": False},
            "__env__": "base",
        }
    }


@pytest.fixture
def fake_remove():
    fake_remove_mod = create_autospec(filemod.remove)
    with patch.dict(file.__salt__, {"file.remove": fake_remove_mod}):
        yield fake_remove_mod


# TODO: This file.absent test should be a functional test instead. For now this is probably good enough -W. Werner, 2020-09-15
@pytest.mark.parametrize("mock_mod", ["os.path.isfile", "os.path.isdir"])
def test_file_absent_should_use_force_mode_for_file_remove(fake_remove, mock_mod):
    expected_path = "/some/abspath/foo"
    with patch(mock_mod, autospec=True, return_value=True):
        file.absent(expected_path)

    fake_remove.assert_called_with(expected_path, force=True)


# TODO: This file.matches test should be a functional test instead. For now this is probably good enough -W. Werner, 2020-09-15
def test_file_tidied_for_file_remove(fake_remove):
    patch_is_dir = patch("os.path.isdir", autospec=True, return_value=True)
    patch_os_walk = patch(
        "os.walk",
        autospec=True,
        return_value=[("some root", ("dirs",), ("file1", "file2"))],
    )
    patch_stat = patch("os.stat", autospec=True)
    with patch_os_walk, patch_is_dir, patch_stat as fake_stat:
        fake_stat.return_value.st_atime = 1600356711.1166897
        fake_stat.return_value.st_mode = 33188
        fake_stat.return_value.st_size = 9001  # It's over 9000!

        file.tidied("/some/directory/tree")

    call_root_file1 = f"some root{os.sep}file1"
    call_root_file2 = f"some root{os.sep}file2"
    fake_remove.assert_has_calls([call(call_root_file1), call(call_root_file2)])


# TODO: This file.copy test should be a functional test instead. For now this is probably good enough -W. Werner, 2020-09-15
def test_file_copy_should_use_provided_force_mode_for_file_remove(fake_remove):

    with patch("os.path.lexists", autospec=True, return_value=True), patch(
        "os.path.isfile", autospec=True, return_value=True
    ), patch("os.path.exists", autospec=True, return_value=True), patch.dict(
        file.__opts__, {"user": "somefakeouser"}
    ), patch(
        "salt.states.file._check_user", autospec=True, return_value=False
    ), patch(
        "salt.utils.hashutils.get_hash", autospec=True, return_value=["12345", "54321"]
    ):
        file.copy_("/tmp/foo", source="/tmp/bar", group="fnord", force=True, mode=777)

    fake_remove.assert_called_with("/tmp/foo", force=True)


def test_file_recurse_directory_test():
    salt_dunder = {
        "cp.list_master_dirs": MagicMock(return_value=[]),
        "file.source_list": MagicMock(return_value=("salt://does_not_exist", "")),
    }
    with patch.dict(file.__salt__, salt_dunder):
        ret = file.recurse("/tmp/test", "salt://does_not_exist", saltenv="base")
        assert ret == {
            "changes": {},
            "comment": "The directory 'does_not_exist' does not exist on the salt fileserver in saltenv 'base'",
            "name": "/tmp/test",
            "result": False,
        }
        salt_dunder["cp.list_master_dirs"].assert_called_once_with(
            saltenv="base",
            prefix="does_not_exist/",
        )



"""
   Test cases for salt.states.kapacitor
"""

import pytest

import salt.states.kapacitor as kapacitor
from tests.support.mock import Mock, mock_open, patch


@pytest.fixture
def configure_loader_modules():
    return {kapacitor: {"__opts__": {"test": False}, "__env__": "test"}}


def _present(
    name="testname",
    tick_script="/tmp/script.tick",
    task_type="stream",
    database="testdb",
    retention_policy="default",
    dbrps=None,
    enable=True,
    task=None,
    define_result=True,
    enable_result=True,
    disable_result=True,
    script="testscript",
):
    """
    Run a "kapacitor.present" state after setting up mocks, and return the
    state return value as well as the mocks to make assertions.
    """
    get_mock = Mock(return_value=task)

    if isinstance(define_result, bool):
        define_result = {"success": define_result}
    define_mock = Mock(return_value=define_result)

    if isinstance(enable_result, bool):
        enable_result = {"success": enable_result}
    enable_mock = Mock(return_value=enable_result)

    if isinstance(disable_result, bool):
        disable_result = {"success": disable_result}
    disable_mock = Mock(return_value=disable_result)

    with patch.dict(
        kapacitor.__salt__,
        {
            "kapacitor.get_task": get_mock,
            "kapacitor.define_task": define_mock,
            "kapacitor.enable_task": enable_mock,
            "kapacitor.disable_task": disable_mock,
        },
    ):
        with patch("salt.utils.files.fopen", mock_open(read_data=script)) as open_mock:
            retval = kapacitor.task_present(
                name,
                tick_script,
                task_type=task_type,
                database=database,
                retention_policy=retention_policy,
                enable=enable,
                dbrps=dbrps,
            )

    return retval, get_mock, define_mock, enable_mock, disable_mock


def _task(
    script="testscript", enabled=True, task_type="stream", db="testdb", rp="default"
):
    return {
        "script": script,
        "enabled": enabled,
        "type": task_type,
        "dbrps": [{"db": db, "rp": rp}],
    }


def test_task_present_new_task():
    ret, get_mock, define_mock, enable_mock, _ = _present(dbrps=["testdb2.default_rp"])
    get_mock.assert_called_once_with("testname")
    define_mock.assert_called_once_with(
        "testname",
        "/tmp/script.tick",
        database="testdb",
        retention_policy="default",
        task_type="stream",
        dbrps=["testdb2.default_rp", "testdb.default"],
    )
    enable_mock.assert_called_once_with("testname")
    assert "TICKscript diff" in ret["changes"]
    assert "enabled" in ret["changes"]
    assert ret["changes"]["enabled"]["new"] is True


def test_task_present_existing_task_updated_script():
    ret, get_mock, define_mock, enable_mock, _ = _present(
        task=_task(script="oldscript")
    )
    get_mock.assert_called_once_with("testname")
    define_mock.assert_called_once_with(
        "testname",
        "/tmp/script.tick",
        database="testdb",
        retention_policy="default",
        task_type="stream",
        dbrps=["testdb.default"],
    )
    assert enable_mock.called is False
    assert "TICKscript diff" in ret["changes"]
    assert "enabled" not in ret["changes"]


def test_task_present_existing_task_not_enabled():
    ret, get_mock, define_mock, enable_mock, _ = _present(task=_task(enabled=False))
    get_mock.assert_called_once_with("testname")
    assert define_mock.called is False
    enable_mock.assert_called_once_with("testname")
    assert "diff" not in ret["changes"]
    assert "enabled" in ret["changes"]
    assert ret["changes"]["enabled"]["new"] is True


def test_task_present_disable_existing_task():
    ret, get_mock, define_mock, _, disable_mock = _present(task=_task(), enable=False)
    get_mock.assert_called_once_with("testname")
    assert define_mock.called is False
    disable_mock.assert_called_once_with("testname")
    assert "diff" not in ret["changes"]
    assert "enabled" in ret["changes"]
    assert ret["changes"]["enabled"]["new"] is False



"""
Zookeeper State

:depends:  kazoo
:configuration: See :py:mod:`salt.modules.zookeeper` for setup instructions.

ACLS
~~~~

For more information about acls, please checkout the kazoo documentation.

http://kazoo.readthedocs.io/en/latest/api/security.html#kazoo.security.make_digest_acl

The following options can be included in the acl dictionary:

    :param username: Username to use for the ACL.
    :param password: A plain-text password to hash.
    :param write: Write permission.
    :type write: bool
    :param create: Create permission.
    :type create: bool
    :param delete: Delete permission.
    :type delete: bool
    :param admin: Admin permission.
    :type admin: bool
    :param all: All permissions.
    :type all: bool
"""

__virtualname__ = "zookeeper"


def __virtual__():
    if "zookeeper.create" in __salt__:
        return __virtualname__
    return (False, "zookeeper module could not be loaded")


def _check_acls(left, right):
    first = not bool(set(left) - set(right))
    second = not bool(set(right) - set(left))
    return first and second


def present(
    name,
    value,
    acls=None,
    ephemeral=False,
    sequence=False,
    makepath=False,
    version=-1,
    profile=None,
    hosts=None,
    scheme=None,
    username=None,
    password=None,
    default_acl=None,
):
    """
    Make sure znode is present in the correct state with the correct acls

    name
        path to znode

    value
        value znode should be set to

    acls
        list of acl dictionaries to set on znode (make sure the ones salt is connected with are included)
        Default: None

    ephemeral
        Boolean to indicate if ephemeral znode should be created
        Default: False

    sequence
        Boolean to indicate if znode path is suffixed with a unique index
        Default: False

    makepath
        Boolean to indicate if the parent paths should be created
        Default: False

    version
        For updating, specify the version which should be updated
        Default: -1 (always match)

    profile
        Configured Zookeeper profile to authenticate with (Default: None)

    hosts
        Lists of Zookeeper Hosts (Default: '127.0.0.1:2181)

    scheme
        Scheme to authenticate with (Default: 'digest')

    username
        Username to authenticate (Default: None)

    password
        Password to authenticate (Default: None)

    default_acl
        Default acls to assign if a node is created in this connection (Default: None)

    .. code-block:: yaml

        add znode:
          zookeeper.present:
            - name: /test/name
            - value: gtmanfred
            - makepath: True

        update znode:
          zookeeper.present:
            - name: /test/name
            - value: daniel
            - acls:
              - username: daniel
                password: test
                read: true
              - username: gtmanfred
                password: test
                read: true
                write: true
                create: true
                delete: true
                admin: true
            - makepath: True
    """

    ret = {
        "name": name,
        "result": False,
        "comment": f"Failed to setup znode {name}",
        "changes": {},
    }
    connkwargs = {
        "profile": profile,
        "hosts": hosts,
        "scheme": scheme,
        "username": username,
        "password": password,
        "default_acl": default_acl,
    }
    if acls is None:
        chk_acls = []
    else:
        chk_acls = [__salt__["zookeeper.make_digest_acl"](**acl) for acl in acls]
    if __salt__["zookeeper.exists"](name, **connkwargs):
        cur_value = __salt__["zookeeper.get"](name, **connkwargs)
        cur_acls = __salt__["zookeeper.get_acls"](name, **connkwargs)
        if cur_value == value and _check_acls(cur_acls, chk_acls):
            ret["result"] = True
            ret["comment"] = (
                "Znode {} is already set to the correct value with the correct acls".format(
                    name
                )
            )
            return ret
        elif __opts__["test"] is True:
            ret["result"] = None
            ret["comment"] = f"Znode {name} is will be updated"
            ret["changes"]["old"] = {}
            ret["changes"]["new"] = {}
            if value != cur_value:
                ret["changes"]["old"]["value"] = cur_value
                ret["changes"]["new"]["value"] = value
            if not _check_acls(chk_acls, cur_acls):
                ret["changes"]["old"]["acls"] = cur_acls
                ret["changes"]["new"]["acls"] = chk_acls
            return ret
        else:
            value_result, acl_result = True, True
            changes = {}
            if value != cur_value:
                __salt__["zookeeper.set"](name, value, version, **connkwargs)
                new_value = __salt__["zookeeper.get"](name, **connkwargs)
                value_result = new_value == value
                changes.setdefault("new", {}).setdefault("value", new_value)
                changes.setdefault("old", {}).setdefault("value", cur_value)
            if chk_acls and not _check_acls(chk_acls, cur_acls):
                __salt__["zookeeper.set_acls"](name, acls, version, **connkwargs)
                new_acls = __salt__["zookeeper.get_acls"](name, **connkwargs)
                acl_result = _check_acls(new_acls, chk_acls)
                changes.setdefault("new", {}).setdefault("acls", new_acls)
                changes.setdefault("old", {}).setdefault("value", cur_acls)
            ret["changes"] = changes
            if value_result and acl_result:
                ret["result"] = True
                ret["comment"] = f"Znode {name} successfully updated"
            return ret

    if __opts__["test"] is True:
        ret["result"] = None
        ret["comment"] = f"{name} is will be created"
        ret["changes"]["old"] = {}
        ret["changes"]["new"] = {}
        ret["changes"]["new"]["acls"] = chk_acls
        ret["changes"]["new"]["value"] = value
        return ret

    __salt__["zookeeper.create"](
        name, value, acls, ephemeral, sequence, makepath, **connkwargs
    )

    value_result, acl_result = True, True
    changes = {"old": {}}

    new_value = __salt__["zookeeper.get"](name, **connkwargs)
    value_result = new_value == value
    changes.setdefault("new", {}).setdefault("value", new_value)

    new_acls = __salt__["zookeeper.get_acls"](name, **connkwargs)
    acl_result = acls is None or _check_acls(new_acls, chk_acls)
    changes.setdefault("new", {}).setdefault("acls", new_acls)

    ret["changes"] = changes
    if value_result and acl_result:
        ret["result"] = True
        ret["comment"] = f"Znode {name} successfully created"

    return ret


def absent(
    name,
    version=-1,
    recursive=False,
    profile=None,
    hosts=None,
    scheme=None,
    username=None,
    password=None,
    default_acl=None,
):
    """
    Make sure znode is absent

    name
        path to znode

    version
        Specify the version which should be deleted
        Default: -1 (always match)

    recursive
        Boolean to indicate if children should be recursively deleted
        Default: False

    profile
        Configured Zookeeper profile to authenticate with (Default: None)

    hosts
        Lists of Zookeeper Hosts (Default: '127.0.0.1:2181)

    scheme
        Scheme to authenticate with (Default: 'digest')

    username
        Username to authenticate (Default: None)

    password
        Password to authenticate (Default: None)

    default_acl
        Default acls to assign if a node is created in this connection (Default: None)

    .. code-block:: yaml

        delete znode:
          zookeeper.absent:
            - name: /test
            - recursive: True
    """
    ret = {
        "name": name,
        "result": False,
        "comment": f"Failed to delete znode {name}",
        "changes": {},
    }
    connkwargs = {
        "profile": profile,
        "hosts": hosts,
        "scheme": scheme,
        "username": username,
        "password": password,
        "default_acl": default_acl,
    }

    if __salt__["zookeeper.exists"](name, **connkwargs) is False:
        ret["result"] = True
        ret["comment"] = f"Znode {name} does not exist"
        return ret

    changes = {}
    changes["value"] = __salt__["zookeeper.get"](name, **connkwargs)
    changes["acls"] = __salt__["zookeeper.get_acls"](name, **connkwargs)
    if recursive is True:
        changes["children"] = __salt__["zookeeper.get_children"](name, **connkwargs)

    if __opts__["test"] is True:
        ret["result"] = None
        ret["comment"] = f"Znode {name} will be removed"
        ret["changes"]["old"] = changes
        return ret

    __salt__["zookeeper.delete"](name, version, recursive, **connkwargs)

    if __salt__["zookeeper.exists"](name, **connkwargs) is False:
        ret["result"] = True
        ret["comment"] = f"Znode {name} has been removed"
        ret["changes"]["old"] = changes

    return ret


def acls(
    name,
    acls,
    version=-1,
    profile=None,
    hosts=None,
    scheme=None,
    username=None,
    password=None,
    default_acl=None,
):
    """
    Update acls on a znode

    name
        path to znode

    acls
        list of acl dictionaries to set on znode

    version
        Specify the version which should be deleted
        Default: -1 (always match)

    profile
        Configured Zookeeper profile to authenticate with (Default: None)

    hosts
        Lists of Zookeeper Hosts (Default: '127.0.0.1:2181)

    scheme
        Scheme to authenticate with (Default: 'digest')

    username
        Username to authenticate (Default: None)

    password
        Password to authenticate (Default: None)

    default_acl
        Default acls to assign if a node is created in this connection (Default: None)

    .. code-block:: yaml

        update acls:
          zookeeper.acls:
            - name: /test/name
            - acls:
              - username: daniel
                password: test
                all: True
              - username: gtmanfred
                password: test
                all: True
    """
    ret = {
        "name": name,
        "result": False,
        "comment": f"Failed to set acls on znode {name}",
        "changes": {},
    }
    connkwargs = {
        "profile": profile,
        "hosts": hosts,
        "scheme": scheme,
        "username": username,
        "password": password,
        "default_acl": default_acl,
    }
    if isinstance(acls, dict):
        acls = [acls]
    chk_acls = [__salt__["zookeeper.make_digest_acl"](**acl) for acl in acls]

    if not __salt__["zookeeper.exists"](name, **connkwargs):
        ret["comment"] += ": Znode does not exist"
        return ret

    cur_acls = __salt__["zookeeper.get_acls"](name, **connkwargs)
    if _check_acls(cur_acls, chk_acls):
        ret["result"] = True
        ret["comment"] = f"Znode {name} acls already set"
        return ret

    if __opts__["test"] is True:
        ret["result"] = None
        ret["comment"] = f"Znode {name} acls will be updated"
        ret["changes"]["old"] = cur_acls
        ret["changes"]["new"] = chk_acls
        return ret

    __salt__["zookeeper.set_acls"](name, acls, version, **connkwargs)

    new_acls = __salt__["zookeeper.get_acls"](name, **connkwargs)
    ret["changes"] = {"old": cur_acls, "new": new_acls}
    if _check_acls(new_acls, chk_acls):
        ret["result"] = True
        ret["comment"] = f"Znode {name} acls updated"
        return ret
    ret["comment"] = f"Znode {name} acls failed to update"
    return ret



import logging

import pytest

import salt.modules.dockermod as docker_mod
from tests.support.mock import MagicMock, patch

log = logging.getLogger(__name__)


@pytest.fixture
def configure_loader_modules():
    return {
        docker_mod: {
            "__utils__": {
                "state.get_sls_opts": MagicMock(
                    return_value={
                        "pillarenv": MagicMock(),
                        "pillar": {},
                        "grains": {},
                    }
                ),
                "args.clean_kwargs": lambda **x: x,
            },
            "__salt__": {
                "config.option": MagicMock(return_value=None),
                "cmd.run": fake_run,
            },
            "__opts__": {"id": "dockermod-unit-test"},
        },
    }


def fake_run(*args, **kwargs):
    log.debug("Fake run call args: %s, kwargs: %s", args, kwargs)
    return "{}"


def test_trans_tar_should_have_grains_in_sls_opts_including_pillar_override():
    container_name = "fnord"
    expected_grains = {
        "roscivs": "bottia",
        "fnord": "dronf",
        "salt": "NaCl",
    }
    expected_pillars = {
        "this": {"is": {"my": {"pillar": "data"}}},
    }
    extra_pillar_data = {"some": "extras"}
    fake_trans_tar = MagicMock(return_value=b"hi")
    patch_trans_tar = patch(
        "salt.modules.dockermod._prepare_trans_tar",
        fake_trans_tar,
    )
    patch_call = patch(
        "salt.modules.dockermod.call",
        MagicMock(return_value=expected_grains),
    )
    fake_get_pillar = MagicMock()
    fake_get_pillar.compile_pillar.return_value = expected_pillars
    patch_pillar = patch(
        "salt.modules.dockermod.salt.pillar.get_pillar",
        MagicMock(return_value=fake_get_pillar),
    )
    patch_run_all = patch(
        "salt.modules.dockermod.run_all",
        MagicMock(return_value={"retcode": 1, "stderr": "early exit test"}),
    )
    with patch_trans_tar, patch_call, patch_pillar, patch_run_all:
        docker_mod.sls(container_name, pillar=extra_pillar_data)
        # TODO: It would be fine if we could make this test require less magic numbers -W. Werner, 2019-08-27
        actual_sls_opts = fake_trans_tar.call_args[0][1]
        for (
            key,
            value,
        ) in expected_grains.items():
            assert key in actual_sls_opts["grains"]
            assert value == actual_sls_opts["grains"][key]
        expected_pillars.update(extra_pillar_data)
        for (
            key,
            value,
        ) in expected_pillars.items():
            assert key in actual_sls_opts["pillar"]
            assert value == actual_sls_opts["pillar"][key]



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>

    Test cases for salt.modules.pam
"""

import pytest

import salt.modules.pam as pam
from tests.support.mock import mock_open, patch


@pytest.fixture
def MOCK_FILE():
    return "ok ok ignore"


@pytest.fixture
def configure_loader_modules():
    return {pam: {}}


def test_read_file(MOCK_FILE):
    """
    Test if the parsing function works
    """
    with patch("os.path.exists", return_value=True), patch(
        "salt.utils.files.fopen", mock_open(read_data=MOCK_FILE)
    ):
        assert pam.read_file("/etc/pam.d/login") == [
            {
                "arguments": [],
                "control_flag": "ok",
                "interface": "ok",
                "module": "ignore",
            }
        ]



"""
This module have been tested on infoblox API v1.2.1,
other versions of the API are likly workable.

:depends: libinfoblox, https://github.com/steverweber/libinfoblox

    libinfoblox can be installed using `pip install libinfoblox`

API documents can be found on your infoblox server at:

    https://INFOBLOX/wapidoc

:configuration: The following configuration defaults can be
    defined (pillar or config files '/etc/salt/master.d/infoblox.conf'):

    .. code-block:: python

        infoblox.config:
            api_sslverify: True
            api_url: 'https://INFOBLOX/wapi/v1.2.1'
            api_user: 'username'
            api_key: 'password'

    Many of the functions accept `api_opts` to override the API config.

    .. code-block:: bash

        salt-call infoblox.get_host name=my.host.com \
            api_url: 'https://INFOBLOX/wapi/v1.2.1' \
            api_user=admin \
            api_key=passs

"""

import time

IMPORT_ERR = None
try:
    import libinfoblox
except Exception as exc:  # pylint: disable=broad-except
    IMPORT_ERR = str(exc)
__virtualname__ = "infoblox"


def __virtual__():
    return (IMPORT_ERR is None, IMPORT_ERR)


cache = {}


def _get_config(**api_opts):
    """
    Return configuration
    user passed api_opts override salt config.get vars
    """
    config = {
        "api_sslverify": True,
        "api_url": "https://INFOBLOX/wapi/v1.2.1",
        "api_user": "",
        "api_key": "",
    }
    if "__salt__" in globals():
        config_key = f"{__virtualname__}.config"
        config.update(__salt__["config.get"](config_key, {}))
    # pylint: disable=C0201
    for k in set(config.keys()) & set(api_opts.keys()):
        config[k] = api_opts[k]
    return config


def _get_infoblox(**api_opts):
    config = _get_config(**api_opts)
    # TODO: perhaps cache in __opts__
    cache_key = "infoblox_session_{},{},{}".format(
        config["api_url"], config["api_user"], config["api_key"]
    )
    if cache_key in cache:
        timedelta = int(time.time()) - cache[cache_key]["time"]
        if cache[cache_key]["obj"] and timedelta < 60:
            return cache[cache_key]["obj"]
    c = {}
    c["time"] = int(time.time())
    c["obj"] = libinfoblox.Session(
        api_sslverify=config["api_sslverify"],
        api_url=config["api_url"],
        api_user=config["api_user"],
        api_key=config["api_key"],
    )
    cache[cache_key] = c
    return c["obj"]


def diff_objects(obja, objb):
    """
    Diff two complex infoblox objects.
    This is used from salt states to detect changes in objects.

    Using ``func:nextavailableip`` will not cause a diff if the ipaddress is in
    range
    """
    return libinfoblox.diff_obj(obja, objb)


def is_ipaddr_in_ipfunc_range(ipaddr, ipfunc):
    """
    Return true if the ipaddress is in the range of the nextavailableip function

    CLI Example:

    .. code-block:: bash

        salt-call infoblox.is_ipaddr_in_ipfunc_range \
            ipaddr="10.0.2.2" ipfunc="func:nextavailableip:10.0.0.0/8"
    """
    return libinfoblox.is_ipaddr_in_ipfunc_range(ipaddr, ipfunc)


def update_host(name, data, **api_opts):
    """
    Update host record. This is a helper call to update_object.

    Find a hosts ``_ref`` then call update_object with the record data.

    CLI Example:

    .. code-block:: bash

        salt-call infoblox.update_host name=fqdn data={}
    """
    o = get_host(name=name, **api_opts)
    return update_object(objref=o["_ref"], data=data, **api_opts)


def update_object(objref, data, **api_opts):
    """
    Update raw infoblox object. This is a low level api call.

    CLI Example:

    .. code-block:: bash

        salt-call infoblox.update_object objref=[ref_of_object] data={}
    """
    if "__opts__" in globals() and __opts__["test"]:
        return {"Test": f"Would attempt to update object: {objref}"}
    infoblox = _get_infoblox(**api_opts)
    return infoblox.update_object(objref, data)


def delete_object(objref, **api_opts):
    """
    Delete infoblox object. This is a low level api call.

    CLI Example:

    .. code-block:: bash

        salt-call infoblox.delete_object objref=[ref_of_object]
    """
    if "__opts__" in globals() and __opts__["test"]:
        return {"Test": f"Would attempt to delete object: {objref}"}
    infoblox = _get_infoblox(**api_opts)
    return infoblox.delete_object(objref)


def create_object(object_type, data, **api_opts):
    """
    Create raw infoblox object. This is a low level api call.

    CLI Example:

    .. code-block:: bash

        salt-call infoblox.update_object object_type=record:host  data={}
    """
    if "__opts__" in globals() and __opts__["test"]:
        return {"Test": f"Would attempt to create object: {object_type}"}
    infoblox = _get_infoblox(**api_opts)
    return infoblox.create_object(object_type, data)


def get_object(
    objref,
    data=None,
    return_fields=None,
    max_results=None,
    ensure_none_or_one_result=False,
    **api_opts,
):
    """
    Get raw infoblox object. This is a low level api call.

    CLI Example:

    .. code-block:: bash

        salt-call infoblox.get_object objref=[_ref of object]
    """
    if not data:
        data = {}
    infoblox = _get_infoblox(**api_opts)
    return infoblox.get_object(
        objref, data, return_fields, max_results, ensure_none_or_one_result
    )


def create_cname(data, **api_opts):
    """
    Create a cname record.

    CLI Example:

    .. code-block:: bash

        salt-call infoblox.create_cname data={ \
            "comment": "cname to example server", \
            "name": "example.example.com", \
            "zone": "example.com", \
            "view": "Internal", \
            "canonical": "example-ha-0.example.com" \
        }
    """
    infoblox = _get_infoblox(**api_opts)
    host = infoblox.create_cname(data=data)
    return host


def get_cname(name=None, canonical=None, return_fields=None, **api_opts):
    """
    Get CNAME information.

    CLI Examples:

    .. code-block:: bash

        salt-call infoblox.get_cname name=example.example.com
        salt-call infoblox.get_cname canonical=example-ha-0.example.com
    """
    infoblox = _get_infoblox(**api_opts)
    o = infoblox.get_cname(name=name, canonical=canonical, return_fields=return_fields)
    return o


def update_cname(name, data, **api_opts):
    """
    Update CNAME. This is a helper call to update_object.

    Find a CNAME ``_ref`` then call update_object with the record data.

    CLI Example:

    .. code-block:: bash

        salt-call infoblox.update_cname name=example.example.com data="{
                'canonical':'example-ha-0.example.com',
                'use_ttl':true,
                'ttl':200,
                'comment':'Salt managed CNAME'}"
    """
    o = get_cname(name=name, **api_opts)
    if not o:
        raise Exception("CNAME record not found")
    return update_object(objref=o["_ref"], data=data, **api_opts)


def delete_cname(name=None, canonical=None, **api_opts):
    """
    Delete CNAME. This is a helper call to delete_object.

    If record is not found, return True

    CLI Examples:

    .. code-block:: bash

        salt-call infoblox.delete_cname name=example.example.com
        salt-call infoblox.delete_cname canonical=example-ha-0.example.com
    """
    cname = get_cname(name=name, canonical=canonical, **api_opts)
    if cname:
        return delete_object(cname["_ref"], **api_opts)
    return True


def get_host(name=None, ipv4addr=None, mac=None, return_fields=None, **api_opts):
    """
    Get host information

    CLI Examples:

    .. code-block:: bash

        salt-call infoblox.get_host hostname.domain.ca
        salt-call infoblox.get_host ipv4addr=123.123.122.12
        salt-call infoblox.get_host mac=00:50:56:84:6e:ae
    """
    infoblox = _get_infoblox(**api_opts)
    host = infoblox.get_host(
        name=name, mac=mac, ipv4addr=ipv4addr, return_fields=return_fields
    )
    return host


def get_host_advanced(name=None, ipv4addr=None, mac=None, **api_opts):
    """
    Get all host information

    CLI Example:

    .. code-block:: bash

        salt-call infoblox.get_host_advanced hostname.domain.ca
    """
    infoblox = _get_infoblox(**api_opts)
    host = infoblox.get_host_advanced(name=name, mac=mac, ipv4addr=ipv4addr)
    return host


def get_host_domainname(name, domains=None, **api_opts):
    """
    Get host domain name

    If no domains are passed, the hostname is checked for a zone in infoblox,
    if no zone split on first dot.

    If domains are provided, the best match out of the list is returned.

    If none are found the return is None

    dots at end of names are ignored.

    CLI Example:

    .. code-block:: bash

        salt-call uwl.get_host_domainname name=localhost.t.domain.com \
            domains=['domain.com', 't.domain.com.']

        # returns: t.domain.com
    """
    name = name.lower().rstrip(".")
    if not domains:
        data = get_host(name=name, **api_opts)
        if data and "zone" in data:
            return data["zone"].lower()
        else:
            if name.count(".") > 1:
                return name[name.find(".") + 1 :]
            return name
    match = ""
    for d in domains:
        d = d.lower().rstrip(".")
        if name.endswith(d) and len(d) > len(match):
            match = d
    return match if match else None


def get_host_hostname(name, domains=None, **api_opts):
    """
    Get hostname

    If no domains are passed, the hostname is checked for a zone in infoblox,
    if no zone split on first dot.

    If domains are provided, the best match out of the list is truncated from
    the fqdn leaving the hostname.

    If no matching domains are found the fqdn is returned.

    dots at end of names are ignored.

    CLI Examples:

    .. code-block:: bash

        salt-call infoblox.get_host_hostname fqdn=localhost.xxx.t.domain.com \
            domains="['domain.com', 't.domain.com']"
        #returns: localhost.xxx

        salt-call infoblox.get_host_hostname fqdn=localhost.xxx.t.domain.com
        #returns: localhost
    """
    name = name.lower().rstrip(".")
    if not domains:
        return name.split(".")[0]
    domain = get_host_domainname(name, domains, **api_opts)
    if domain and domain in name:
        return name.rsplit("." + domain)[0]
    return name


def get_host_mac(name=None, allow_array=False, **api_opts):
    """
    Get mac address from host record.

    Use `allow_array` to return possible multiple values.

    CLI Example:

    .. code-block:: bash

        salt-call infoblox.get_host_mac host=localhost.domain.com
    """
    data = get_host(name=name, **api_opts)
    if data and "ipv4addrs" in data:
        l = []
        for a in data["ipv4addrs"]:
            if "mac" in a:
                l.append(a["mac"])
        if allow_array:
            return l
        if l:
            return l[0]
    return None


def get_host_ipv4(name=None, mac=None, allow_array=False, **api_opts):
    """
    Get ipv4 address from host record.

    Use `allow_array` to return possible multiple values.

    CLI Examples:

    .. code-block:: bash

        salt-call infoblox.get_host_ipv4 host=localhost.domain.com
        salt-call infoblox.get_host_ipv4 mac=00:50:56:84:6e:ae
    """
    data = get_host(name=name, mac=mac, **api_opts)
    if data and "ipv4addrs" in data:
        l = []
        for a in data["ipv4addrs"]:
            if "ipv4addr" in a:
                l.append(a["ipv4addr"])
        if allow_array:
            return l
        if l:
            return l[0]
    return None


def get_host_ipv4addr_info(
    ipv4addr=None, mac=None, discovered_data=None, return_fields=None, **api_opts
):
    """
    Get host ipv4addr information

    CLI Examples:

    .. code-block:: bash

        salt-call infoblox.get_ipv4addr ipv4addr=123.123.122.12
        salt-call infoblox.get_ipv4addr mac=00:50:56:84:6e:ae
        salt-call infoblox.get_ipv4addr mac=00:50:56:84:6e:ae return_fields=host return_fields='mac,host,configure_for_dhcp,ipv4addr'
    """
    infoblox = _get_infoblox(**api_opts)
    return infoblox.get_host_ipv4addr_object(
        ipv4addr, mac, discovered_data, return_fields
    )


def get_host_ipv6addr_info(
    ipv6addr=None, mac=None, discovered_data=None, return_fields=None, **api_opts
):
    """
    Get host ipv6addr information

    CLI Example:

    .. code-block:: bash

        salt-call infoblox.get_host_ipv6addr_info ipv6addr=2001:db8:85a3:8d3:1349:8a2e:370:7348
    """
    infoblox = _get_infoblox(**api_opts)
    return infoblox.get_host_ipv6addr_object(
        ipv6addr, mac, discovered_data, return_fields
    )


def get_network(ipv4addr=None, network=None, return_fields=None, **api_opts):
    """
    Get list of all networks. This is helpful when looking up subnets to use
    with func:nextavailableip

    This call is offen slow and not cached!

    some return_fields
    comment,network,network_view,ddns_domainname,disable,enable_ddns

    CLI Example:

    .. code-block:: bash

        salt-call infoblox.get_network
    """
    infoblox = _get_infoblox(**api_opts)
    return infoblox.get_network(
        ipv4addr=ipv4addr, network=network, return_fields=return_fields
    )


def delete_host(name=None, mac=None, ipv4addr=None, **api_opts):
    """
    Delete host

    CLI Example:

    .. code-block:: bash

        salt-call infoblox.delete_host name=example.domain.com
        salt-call infoblox.delete_host ipv4addr=123.123.122.12
        salt-call infoblox.delete_host ipv4addr=123.123.122.12 mac=00:50:56:84:6e:ae
    """
    if "__opts__" in globals() and __opts__["test"]:
        return {"Test": "Would attempt to delete host"}
    infoblox = _get_infoblox(**api_opts)
    return infoblox.delete_host(name, mac, ipv4addr)


def create_host(data, **api_opts):
    """
    Add host record

    Avoid race conditions, use func:nextavailableip for ipv[4,6]addrs:

    - func:nextavailableip:network/ZG54dfgsrDFEFfsfsLzA:10.0.0.0/8/default
    - func:nextavailableip:10.0.0.0/8
    - func:nextavailableip:10.0.0.0/8,external
    - func:nextavailableip:10.0.0.3-10.0.0.10

    See your infoblox API for full `data` format.

    CLI Example:

    .. code-block:: bash

        salt-call infoblox.create_host \
            data =
                {'name': 'hostname.example.ca',
                'aliases': ['hostname.math.example.ca'],
            'extattrs': [{'Business Contact': {'value': 'example@example.ca'}},
                {'Pol8 Classification': {'value': 'Restricted'}},
                {'Primary OU': {'value': 'CS'}},
                {'Technical Contact': {'value': 'example@example.ca'}}],
            'ipv4addrs': [{'configure_for_dhcp': True,
                'ipv4addr': 'func:nextavailableip:129.97.139.0/24',
                'mac': '00:50:56:84:6e:ae'}],
            'ipv6addrs': [], }
    """
    return create_object("record:host", data, **api_opts)


def get_ipv4_range(start_addr=None, end_addr=None, return_fields=None, **api_opts):
    """
    Get ip range

    CLI Example:

    .. code-block:: bash

        salt-call infoblox.get_ipv4_range start_addr=123.123.122.12
    """
    infoblox = _get_infoblox(**api_opts)
    return infoblox.get_range(start_addr, end_addr, return_fields)


def delete_ipv4_range(start_addr=None, end_addr=None, **api_opts):
    """
    Delete ip range.

    CLI Example:

    .. code-block:: bash

        salt-call infoblox.delete_ipv4_range start_addr=123.123.122.12
    """
    r = get_ipv4_range(start_addr, end_addr, **api_opts)
    if r:
        return delete_object(r["_ref"], **api_opts)
    else:
        return True


def create_ipv4_range(data, **api_opts):
    """
    Create a ipv4 range

    This is a helper function to `create_object`
    See your infoblox API for full `data` format.

    CLI Example:

    .. code-block:: bash

        salt-call infoblox.create_ipv4_range data={
            start_addr: '129.97.150.160',
            end_addr: '129.97.150.170'}
    """
    return create_object("range", data, **api_opts)


def create_a(data, **api_opts):
    """
    Create A record.

    This is a helper function to `create_object`.
    See your infoblox API for full `data` format.

    CLI Example:

    .. code-block:: bash

        salt-call infoblox.create_a \
                    data =
                    name: 'fastlinux.math.example.ca'
                    ipv4addr: '127.0.0.1'
                    view: External
    """
    return create_object("record:a", data, **api_opts)


def get_a(name=None, ipv4addr=None, allow_array=True, **api_opts):
    """
    Get A record

    CLI Examples:

    .. code-block:: bash

        salt-call infoblox.get_a name=abc.example.com
        salt-call infoblox.get_a ipv4addr=192.168.3.5
    """
    data = {}
    if name:
        data["name"] = name
    if ipv4addr:
        data["ipv4addr"] = ipv4addr
    r = get_object("record:a", data=data, **api_opts)
    if r and len(r) > 1 and not allow_array:
        raise Exception("More than one result, use allow_array to return the data")
    return r


def delete_a(name=None, ipv4addr=None, allow_array=False, **api_opts):
    """
    Delete A record

    If the A record is used as a round robin you can set ``allow_array=True`` to
    delete all records for the hostname.

    CLI Examples:

    .. code-block:: bash

        salt-call infoblox.delete_a name=abc.example.com
        salt-call infoblox.delete_a ipv4addr=192.168.3.5
        salt-call infoblox.delete_a name=acname.example.com allow_array=True
    """
    r = get_a(name, ipv4addr, allow_array=False, **api_opts)
    if not r:
        return True
    if len(r) > 1 and not allow_array:
        raise Exception("More than one result, use allow_array to override")
    ret = []
    for ri in r:
        ret.append(delete_object(ri["_ref"], **api_opts))
    return ret



import pytest


@pytest.fixture
def saltcheck(modules):
    return modules.saltcheck


@pytest.mark.slow_test
def test_saltcheck_render_pyobjects_state(state_tree, saltcheck):
    with pytest.helpers.temp_file("pyobj_touched.txt") as tpath:
        sls_content = f"""
        #!pyobjects

        File.touch("{tpath}")
        """
        tst_content = f"""
        is_stuff_there:
          module_and_function: file.file_exists
          args:
            - "{tpath}"
          assertion: assertTrue
        """
        with pytest.helpers.temp_file(
            "pyobj_touched/init.sls", sls_content, state_tree
        ), pytest.helpers.temp_file(
            "pyobj_touched/saltcheck-tests/init.tst", tst_content, state_tree
        ):
            ret = saltcheck.run_state_tests("pyobj_touched")
            assert ret[0]["pyobj_touched"]["is_stuff_there"]["status"] == "Pass"
            assert ret[1]["TEST RESULTS"]["Passed"] == 1
            assert ret[1]["TEST RESULTS"]["Missing Tests"] == 0
            assert ret[1]["TEST RESULTS"]["Failed"] == 0
            assert ret[1]["TEST RESULTS"]["Skipped"] == 0


@pytest.mark.slow_test
def test_saltcheck_allow_remote_fileclient(state_tree, saltcheck):
    sls_content = """
    test_state:
      test.show_notification:
        - text: The test state
    """

    tst_content = """
    test cp.cache_file:
      module_and_function: cp.cache_file
      args:
        - salt://sltchk_remote/download_me.txt
      kwargs:
        saltenv: base
      assertion: assertNotEmpty
      output_details: True
    """

    with pytest.helpers.temp_file(
        "sltchk_remote/init.sls", sls_content, state_tree
    ), pytest.helpers.temp_file(
        "sltchk_remote/saltcheck-tests/init.tst", tst_content, state_tree
    ), pytest.helpers.temp_file(
        "sltchk_remote/download_me.txt", "salty", state_tree
    ):

        ret = saltcheck.run_state_tests("sltchk_remote")
        assert ret[0]["sltchk_remote"]["test cp.cache_file"]["status"] == "Pass"
        assert ret[1]["TEST RESULTS"]["Passed"] == 1
        assert ret[1]["TEST RESULTS"]["Missing Tests"] == 0
        assert ret[1]["TEST RESULTS"]["Failed"] == 0
        assert ret[1]["TEST RESULTS"]["Skipped"] == 0



import os
import sys

import pytest

import salt.modules.ini_manage as ini
import salt.utils.files
import salt.utils.stringutils


@pytest.fixture
def ini_content():
    return [
        "# Comment on the first line",
        "",
        "# First main option",
        "option1=main1",
        "",
        "# Second main option",
        "option2=main2",
        "",
        "",
        "[main]",
        "# Another comment",
        "test1=value 1",
        "",
        "test2=value 2",
        "",
        "[SectionB]",
        "test1=value 1B",
        "",
        "# Blank line should be above",
        "test3 = value 3B",
        "",
        "[SectionC]",
        "# The following option is empty",
        "empty_option=",
    ]


@pytest.fixture(scope="function")
def ini_file(tmp_path, ini_content):
    file_path = tmp_path / "file.ini"
    yield file_path


@pytest.fixture
def unicode_content():
    return [
        "# An ini file with some unicode characters",
        "",
        "[Ascii]",
        "de = Deutsch",
        "en_GB = English (UK)",
        "en_US = English (US)",
        "fi = Suomi",
        "hu = Magyar",
        "it = Italiano",
        "nl = Dutch",
        "pt = Portuguese",
        "sv = Svenska",
        "",
        "[Юникод]",
        "# This means unicode in Russian",
        "es = Español",
        "es_ES = Español (ES)",
        "fr = Français",
        "hi = हिंदी",
        "ja = 日本語",
        "ko = :한국어",
        "zh = 简体中文",
        "繁體中文 = zh_TW",
    ]


def test_section_req():
    """
    Test the __repr__ in the _Section class
    """
    expected = f"_Section(){os.linesep}{{}}"
    assert repr(ini._Section("test")) == expected


@pytest.mark.parametrize("linesep", ["\r", "\n", "\r\n"])
@pytest.mark.parametrize(
    "encoding", [None, "cp1252" if sys.platform == "win32" else "ISO-2022-JP"]
)
def test_get_option(encoding, linesep, ini_file, ini_content):
    """
    Test get_option method.
    """
    content = salt.utils.stringutils.to_bytes(
        linesep.join(ini_content), encoding=encoding
    )
    ini_file.write_bytes(content)

    option = ini.get_option(str(ini_file), "main", "test1", encoding=encoding)
    assert option == "value 1"

    option = ini.get_option(str(ini_file), "main", "test2", encoding=encoding)
    assert option == "value 2"

    option = ini.get_option(str(ini_file), "SectionB", "test1", encoding=encoding)
    assert option == "value 1B"

    option = ini.get_option(str(ini_file), "SectionB", "test3", encoding=encoding)
    assert option == "value 3B"

    option = ini.get_option(
        str(ini_file), "SectionC", "empty_option", encoding=encoding
    )
    assert option == ""


@pytest.mark.parametrize("linesep", ["\r", "\n", "\r\n"])
@pytest.mark.parametrize(
    "encoding", [None, "cp1252" if sys.platform == "win32" else "ISO-2022-JP"]
)
def test_get_section(encoding, linesep, ini_file, ini_content):
    """
    Test get_section method.
    """
    content = salt.utils.stringutils.to_bytes(
        linesep.join(ini_content), encoding=encoding
    )
    ini_file.write_bytes(content)

    expected = {"test1": "value 1B", "test3": "value 3B"}
    assert ini.get_section(str(ini_file), "SectionB", encoding=encoding) == expected


@pytest.mark.parametrize("linesep", ["\r", "\n", "\r\n"])
@pytest.mark.parametrize(
    "encoding", [None, "cp1252" if sys.platform == "win32" else "ISO-2022-JP"]
)
def test_remove_option(encoding, linesep, ini_file, ini_content):
    """
    Test remove_option method.
    """
    content = salt.utils.stringutils.to_bytes(
        linesep.join(ini_content), encoding=encoding
    )
    ini_file.write_bytes(content)

    assert (
        ini.remove_option(str(ini_file), "SectionB", "test1", encoding=encoding)
        == "value 1B"
    )
    assert ini.get_option(str(ini_file), "SectionB", "test1", encoding=encoding) is None


@pytest.mark.parametrize("linesep", ["\r", "\n", "\r\n"])
@pytest.mark.parametrize(
    "encoding", [None, "cp1252" if sys.platform == "win32" else "ISO-2022-JP"]
)
def test_remove_section(encoding, linesep, ini_file, ini_content):
    """
    Test remove_section method.
    """
    content = salt.utils.stringutils.to_bytes(
        linesep.join(ini_content), encoding=encoding
    )
    ini_file.write_bytes(content)

    expected = {"test1": "value 1B", "test3": "value 3B"}
    assert ini.remove_section(str(ini_file), "SectionB", encoding=encoding) == expected
    assert ini.get_section(str(ini_file), "SectionB", encoding=encoding) == {}


@pytest.mark.parametrize("linesep", ["\r", "\n", "\r\n"])
@pytest.mark.parametrize(
    "encoding", [None, "cp1252" if sys.platform == "win32" else "ISO-2022-JP"]
)
def test_get_ini(encoding, linesep, ini_file, ini_content):
    """
    Test get_ini method.
    """
    content = salt.utils.stringutils.to_bytes(
        linesep.join(ini_content), encoding=encoding
    )
    ini_file.write_bytes(content)

    expected = {
        "SectionC": {
            "empty_option": "",
        },
        "SectionB": {
            "test1": "value 1B",
            "test3": "value 3B",
        },
        "main": {
            "test1": "value 1",
            "test2": "value 2",
        },
        "option2": "main2",
        "option1": "main1",
    }
    assert dict(ini.get_ini(str(ini_file), encoding=encoding)) == expected


@pytest.mark.parametrize("linesep", ["\r", "\n", "\r\n"])
@pytest.mark.parametrize(
    "encoding", [None, "cp1252" if sys.platform == "win32" else "ISO-2022-JP"]
)
def test_set_option(encoding, linesep, ini_file, ini_content):
    """
    Test set_option method.
    """
    content = salt.utils.stringutils.to_bytes(
        linesep.join(ini_content), encoding=encoding
    )
    ini_file.write_bytes(content)

    result = ini.set_option(
        str(ini_file),
        {
            "SectionB": {
                "test3": "new value 3B",
                "test_set_option": "test_set_value",
            },
            "SectionD": {"test_set_option2": "test_set_value1"},
        },
        encoding=encoding,
    )
    expected = {
        "SectionB": {
            "test3": {"after": "new value 3B", "before": "value 3B"},
            "test_set_option": {"after": "test_set_value", "before": None},
        },
        "SectionD": {
            "after": {"test_set_option2": "test_set_value1"},
            "before": None,
        },
    }
    assert result == expected

    # Check existing option updated
    assert (
        ini.get_option(str(ini_file), "SectionB", "test3", encoding=encoding)
        == "new value 3B"
    )

    # Check new section and option added
    assert (
        ini.get_option(str(ini_file), "SectionD", "test_set_option2", encoding=encoding)
        == "test_set_value1"
    )


@pytest.mark.parametrize("no_spaces", [True, False])
@pytest.mark.parametrize("linesep", ["\r", "\n", "\r\n"])
@pytest.mark.parametrize(
    "encoding", [None, "cp1252" if sys.platform == "win32" else "ISO-2022-JP"]
)
def test_empty_value(encoding, linesep, no_spaces, ini_file, ini_content):
    """
    Test empty value preserved after edit
    """
    content = salt.utils.stringutils.to_bytes(
        linesep.join(ini_content), encoding=encoding
    )
    ini_file.write_bytes(content)

    ini.set_option(
        str(ini_file),
        {"SectionB": {"test3": "new value 3B"}},
        encoding=encoding,
        no_spaces=no_spaces,
    )
    with salt.utils.files.fopen(str(ini_file), "r") as fp_:
        file_content = salt.utils.stringutils.to_unicode(fp_.read(), encoding=encoding)
    expected = f"{os.linesep}empty_option{'=' if no_spaces else ' = '}{os.linesep}"
    assert expected in file_content, "empty_option was not preserved"


@pytest.mark.parametrize("no_spaces", [True, False])
@pytest.mark.parametrize("linesep", ["\r", "\n", "\r\n"])
@pytest.mark.parametrize(
    "encoding", [None, "cp1252" if sys.platform == "win32" else "ISO-2022-JP"]
)
def test_empty_lines(encoding, linesep, no_spaces, ini_file, ini_content):
    """
    Test empty lines preserved after edit
    """
    content = salt.utils.stringutils.to_bytes(
        linesep.join(ini_content), encoding=encoding
    )
    ini_file.write_bytes(content)

    expected = os.linesep.join(
        [
            "# Comment on the first line",
            "",
            "# First main option",
            f"option1{'=' if no_spaces else ' = '}main1",
            "",
            "# Second main option",
            f"option2{'=' if no_spaces else ' = '}main2",
            "",
            "[main]",
            "# Another comment",
            f"test1{'=' if no_spaces else ' = '}value 1",
            "",
            f"test2{'=' if no_spaces else ' = '}value 2",
            "",
            "[SectionB]",
            f"test1{'=' if no_spaces else ' = '}value 1B",
            "",
            "# Blank line should be above",
            f"test3{'=' if no_spaces else ' = '}new value 3B",
            "",
            "[SectionC]",
            "# The following option is empty",
            f"empty_option{'=' if no_spaces else ' = '}",
            "",
        ]
    )
    ini.set_option(
        str(ini_file),
        {"SectionB": {"test3": "new value 3B"}},
        encoding=encoding,
        no_spaces=no_spaces,
    )
    with salt.utils.files.fopen(str(ini_file), "r") as fp_:
        file_content = fp_.read()
    assert expected == file_content


@pytest.mark.parametrize("no_spaces", [True, False])
@pytest.mark.parametrize("linesep", ["\r", "\n", "\r\n"])
@pytest.mark.parametrize(
    "encoding", [None, "cp1252" if sys.platform == "win32" else "ISO-2022-JP"]
)
def test_empty_lines_multiple_edits(
    encoding, linesep, no_spaces, ini_file, ini_content
):
    """
    Test empty lines preserved after multiple edits
    """
    content = salt.utils.stringutils.to_bytes(
        linesep.join(ini_content), encoding=encoding
    )
    ini_file.write_bytes(content)

    ini.set_option(
        str(ini_file),
        {"SectionB": {"test3": "this value will be edited two times"}},
        encoding=encoding,
        no_spaces=no_spaces,
    )

    expected = os.linesep.join(
        [
            "# Comment on the first line",
            "",
            "# First main option",
            f"option1{'=' if no_spaces else ' = '}main1",
            "",
            "# Second main option",
            f"option2{'=' if no_spaces else ' = '}main2",
            "",
            "[main]",
            "# Another comment",
            f"test1{'=' if no_spaces else ' = '}value 1",
            "",
            f"test2{'=' if no_spaces else ' = '}value 2",
            "",
            "[SectionB]",
            f"test1{'=' if no_spaces else ' = '}value 1B",
            "",
            "# Blank line should be above",
            f"test3{'=' if no_spaces else ' = '}new value 3B",
            "",
            "[SectionC]",
            "# The following option is empty",
            f"empty_option{'=' if no_spaces else ' = '}",
            "",
        ]
    )
    ini.set_option(
        str(ini_file),
        {"SectionB": {"test3": "new value 3B"}},
        encoding=encoding,
        no_spaces=no_spaces,
    )
    with salt.utils.files.fopen(str(ini_file), "r") as fp_:
        file_content = fp_.read()
    assert expected == file_content


@pytest.mark.parametrize("linesep", ["\r", "\n", "\r\n"])
@pytest.mark.parametrize("encoding", [None, "utf-16", "utf-32-le"])
def test_unicode_get_option(encoding, linesep, ini_file, unicode_content):
    """
    Test ability to get an option from a file that contains unicode characters
    We can't encode the file with something that doesn't support unicode
    Ie: cp1252
    """
    content = salt.utils.stringutils.to_bytes(
        linesep.join(unicode_content), encoding=encoding
    )
    ini_file.write_bytes(content)

    # Get a non-unicode value
    assert ini.get_option(str(ini_file), "Ascii", "de", encoding=encoding) == "Deutsch"

    # Get a unicode value
    assert ini.get_option(str(ini_file), "Юникод", "hi", encoding=encoding) == "हिंदी"


@pytest.mark.parametrize("linesep", ["\r", "\n", "\r\n"])
@pytest.mark.parametrize("encoding", [None, "utf-16", "utf-16-le", "utf-32-le"])
def test_unicode_set_option(encoding, linesep, ini_file, unicode_content):
    """
    Test ability to set an option in a file that contains unicode characters.
    The option itself may be unicode
    We can't encode the file with something that doesn't support unicode
    Ie: cp1252
    """
    content = salt.utils.stringutils.to_bytes(
        linesep.join(unicode_content), encoding=encoding
    )
    ini_file.write_bytes(content)

    result = ini.set_option(
        str(ini_file),
        {
            "Ascii": {"ay": "Aymar"},
            "Юникод": {"dv": "ދިވެހިބަސް"},
        },
        encoding=encoding,
    )
    expected = {
        "Ascii": {
            "ay": {
                "before": None,
                "after": "Aymar",
            },
        },
        "Юникод": {
            "dv": {
                "before": None,
                "after": "ދިވެހިބަސް",
            },
        },
    }
    assert result == expected

    # Check existing option updated
    assert ini.get_option(str(ini_file), "Ascii", "ay", encoding=encoding) == "Aymar"

    # Check new section and option added
    assert ini.get_option(str(ini_file), "Юникод", "dv", encoding=encoding) == "ދިވެހިބަސް"


@pytest.mark.parametrize("linesep", ["\r", "\n", "\r\n"])
@pytest.mark.parametrize("encoding", [None, "utf-16", "utf-16-le", "utf-32-le"])
def test_unicode_get_section(encoding, linesep, ini_file, unicode_content):
    """
    Test get_section method.
    """
    content = salt.utils.stringutils.to_bytes(
        linesep.join(unicode_content), encoding=encoding
    )
    ini_file.write_bytes(content)

    expected = {
        "es": "Español",
        "es_ES": "Español (ES)",
        "fr": "Français",
        "hi": "हिंदी",
        "ja": "日本語",
        "ko": ":한국어",
        "zh": "简体中文",
        "繁體中文": "zh_TW",
    }
    assert ini.get_section(str(ini_file), "Юникод", encoding=encoding) == expected


@pytest.mark.parametrize("linesep", ["\r", "\n", "\r\n"])
@pytest.mark.parametrize("encoding", [None, "utf-16", "utf-16-le", "utf-32-le"])
def test_unicode_remove_option(encoding, linesep, ini_file, unicode_content):
    """
    Test remove_option method.
    """
    content = salt.utils.stringutils.to_bytes(
        linesep.join(unicode_content), encoding=encoding
    )
    ini_file.write_bytes(content)

    assert (
        ini.remove_option(str(ini_file), "Юникод", "繁體中文", encoding=encoding)
        == "zh_TW"
    )
    assert (
        ini.get_option(str(ini_file), "Юникод", "繁體中文", encoding=encoding) is None
    )


@pytest.mark.parametrize("linesep", ["\r", "\n", "\r\n"])
@pytest.mark.parametrize("encoding", [None, "utf-16", "utf-16-le", "utf-32-le"])
def test_unicode_remove_section(encoding, linesep, ini_file, unicode_content):
    """
    Test remove_section method.
    """
    content = salt.utils.stringutils.to_bytes(
        linesep.join(unicode_content), encoding=encoding
    )
    ini_file.write_bytes(content)

    expected = {
        "es": "Español",
        "es_ES": "Español (ES)",
        "fr": "Français",
        "hi": "हिंदी",
        "ja": "日本語",
        "ko": ":한국어",
        "zh": "简体中文",
        "繁體中文": "zh_TW",
    }
    assert ini.remove_section(str(ini_file), "Юникод", encoding=encoding) == expected
    assert ini.get_section(str(ini_file), "Юникод", encoding=encoding) == {}



"""
Management of MySQL users
=========================

:depends:   - MySQLdb Python module
:configuration: See :py:mod:`salt.modules.mysql` for setup instructions.

.. code-block:: yaml

    frank:
      mysql_user.present:
        - host: localhost
        - password: bobcat


.. versionadded:: 0.16.2
    Authentication overrides have been added.

The MySQL authentication information specified in the minion config file can be
overridden in states using the following arguments: ``connection_host``,
``connection_port``, ``connection_user``, ``connection_pass``,
``connection_db``, ``connection_unix_socket``, ``connection_default_file`` and
``connection_charset``.

.. code-block:: yaml

    frank:
      mysql_user.present:
        - host: localhost
        - password: "bob@cat"
        - connection_user: someuser
        - connection_pass: somepass
        - connection_charset: utf8
        - saltenv:
          - LC_ALL: "en_US.utf8"


This state is not able to grant permissions for the user. See
:py:mod:`salt.states.mysql_grants` for further instructions.

"""

import sys

import salt.utils.data


def __virtual__():
    """
    Only load if the mysql module is in __salt__
    """
    if "mysql.user_create" in __salt__:
        return True
    return (False, "mysql module could not be loaded")


def _get_mysql_error():
    """
    Look in module context for a MySQL error. Eventually we should make a less
    ugly way of doing this.
    """
    return sys.modules[__salt__["test.ping"].__module__].__context__.pop(
        "mysql.error", None
    )


def present(
    name,
    host="localhost",
    password=None,
    password_hash=None,
    allow_passwordless=False,
    unix_socket=False,
    password_column=None,
    auth_plugin="mysql_native_password",
    **connection_args,
):
    """
    Ensure that the named user is present with the specified properties. A
    passwordless user can be configured by omitting ``password`` and
    ``password_hash``, and setting ``allow_passwordless`` to ``True``.

    name
        The name of the user to manage

    host
        Host for which this user/password combo applies

    password
        The password to use for this user. Will take precedence over the
        ``password_hash`` option if both are specified.

    password_hash
        The password in hashed form. Be sure to quote the password because YAML
        doesn't like the ``*``. A password hash can be obtained from the mysql
        command-line client like so::

            mysql> SELECT PASSWORD('mypass');
            +-------------------------------------------+
            | PASSWORD('mypass')                        |
            +-------------------------------------------+
            | *6C8989366EAF75BB670AD8EA7A7FC1176A95CEF4 |
            +-------------------------------------------+
            1 row in set (0.00 sec)

    allow_passwordless
        If ``True``, then ``password`` and ``password_hash`` can be omitted to
        permit a passwordless login.

        .. versionadded:: 0.16.2

    unix_socket
        If ``True`` and allow_passwordless is ``True``, the unix_socket auth
        plugin will be used.
    """
    ret = {
        "name": name,
        "changes": {},
        "result": True,
        "comment": f"User {name}@{host} is already present",
    }

    passwordless = not any((password, password_hash))

    # check if user exists with the same password (or passwordless login)
    if passwordless:
        if not salt.utils.data.is_true(allow_passwordless) and not unix_socket:
            ret["comment"] = (
                "Either password or password_hash must be "
                "specified, unless allow_passwordless is True"
            )
            ret["result"] = False
            return ret
        else:
            if __salt__["mysql.user_exists"](
                name,
                host,
                passwordless=True,
                unix_socket=unix_socket,
                password_column=password_column,
                **connection_args,
            ):
                if allow_passwordless:
                    ret["comment"] += " with passwordless login"
                return ret
            else:
                err = _get_mysql_error()
                if err is not None:
                    ret["comment"] = err
                    ret["result"] = False
                    return ret
    else:
        if __salt__["mysql.user_exists"](
            name,
            host,
            password,
            password_hash,
            unix_socket=unix_socket,
            password_column=password_column,
            **connection_args,
        ):
            if auth_plugin == "mysql_native_password":
                ret["comment"] += " with the desired password"
                if password_hash and not password:
                    ret["comment"] += " hash"
            else:
                ret["comment"] += ". Unable to verify password."
            return ret
        else:
            err = _get_mysql_error()
            if err is not None:
                ret["comment"] = err
                ret["result"] = False
                return ret

    # check if user exists with a different password
    if __salt__["mysql.user_exists"](
        name, host, unix_socket=unix_socket, **connection_args
    ):

        # The user is present, change the password
        if __opts__["test"]:
            ret["comment"] = f"Password for user {name}@{host} is set to be "
            ret["result"] = None
            if passwordless:
                ret["comment"] += "cleared"
                if not salt.utils.data.is_true(allow_passwordless):
                    ret["comment"] += ", but allow_passwordless != True"
                    ret["result"] = False
            else:
                ret["comment"] += "changed"
            return ret

        if __salt__["mysql.user_chpass"](
            name,
            host,
            password,
            password_hash,
            allow_passwordless,
            unix_socket,
            **connection_args,
        ):
            ret["comment"] = "Password for user {}@{} has been {}".format(
                name, host, "cleared" if passwordless else "changed"
            )
            ret["changes"][name] = "Updated"
        else:
            ret["comment"] = "Failed to {} password for user {}@{}".format(
                "clear" if passwordless else "change", name, host
            )
            err = _get_mysql_error()
            if err is not None:
                ret["comment"] += f" ({err})"
            if passwordless and not salt.utils.data.is_true(allow_passwordless):
                ret["comment"] += (
                    ". Note: allow_passwordless must be True "
                    "to permit passwordless login."
                )
            ret["result"] = False
    else:

        err = _get_mysql_error()
        if err is not None:
            ret["comment"] = err
            ret["result"] = False
            return ret

        # The user is not present, make it!
        if __opts__["test"]:
            ret["comment"] = f"User {name}@{host} is set to be added"
            ret["result"] = None
            if allow_passwordless:
                ret["comment"] += " with passwordless login"
                if not salt.utils.data.is_true(allow_passwordless):
                    ret["comment"] += ", but allow_passwordless != True"
                    ret["result"] = False
            if unix_socket:
                ret["comment"] += " using unix_socket"
            return ret

        if __salt__["mysql.user_create"](
            name,
            host,
            password,
            password_hash,
            allow_passwordless,
            unix_socket=unix_socket,
            password_column=password_column,
            auth_plugin=auth_plugin,
            **connection_args,
        ):
            ret["comment"] = f"The user {name}@{host} has been added"
            if allow_passwordless:
                ret["comment"] += " with passwordless login"
            if unix_socket:
                ret["comment"] += " using unix_socket"
            ret["changes"][name] = "Present"
        else:
            ret["comment"] = f"Failed to create user {name}@{host}"
            err = _get_mysql_error()
            if err is not None:
                ret["comment"] += f" ({err})"
            ret["result"] = False

    return ret


def absent(name, host="localhost", **connection_args):
    """
    Ensure that the named user is absent

    name
        The name of the user to remove
    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    # Check if user exists, and if so, remove it
    if __salt__["mysql.user_exists"](name, host, **connection_args):
        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = f"User {name}@{host} is set to be removed"
            return ret
        if __salt__["mysql.user_remove"](name, host, **connection_args):
            ret["comment"] = f"User {name}@{host} has been removed"
            ret["changes"][name] = "Absent"
            return ret
        else:
            err = _get_mysql_error()
            if err is not None:
                ret["comment"] = err
                ret["result"] = False
                return ret
    else:
        err = _get_mysql_error()
        if err is not None:
            ret["comment"] = err
            ret["result"] = False
            return ret

    # fallback
    ret["comment"] = "User {}@{} is not present, so it cannot be removed".format(
        name, host
    )
    return ret



"""
    :codeauthor: Rupesh Tare <rupesht@saltstack.com>
"""

import subprocess

import pytest

import salt.modules.localemod as localemod
from salt.exceptions import CommandExecutionError
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, Mock, patch
from tests.support.unit import TestCase


def _check_localectl():
    if not hasattr(_check_localectl, "memo"):
        proc = subprocess.run(["localectl"], check=False, capture_output=True)
        _check_localectl.memo = (
            b"Failed to connect to bus: No such file or directory" in proc.stderr
        )
    return _check_localectl.memo


class LocalemodTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.localemod
    """

    locale_ctl_out = """
   System Locale: LANG=de_DE.utf8
                  LANGUAGE=de_DE.utf8
       VC Keymap: n/a
      X11 Layout: us
       X11 Model: pc105
    """
    locale_ctl_notset = """
       System Locale: n/a

       VC Keymap: n/a
      X11 Layout: n/a
       X11 Model: n/a
    """
    locale_ctl_out_empty = ""
    locale_ctl_out_broken = """
    System error:Recursive traversal of loopback mount points
    """
    locale_ctl_out_structure = """
       Main: printers=We're upgrading /dev/null
             racks=hardware stress fractures
             failure=Ionisation from the air-conditioning
    Cow say: errors=We're out of slots on the server
             hardware=high pressure system failure
     Reason: The vendor put the bug there.
    """

    def setup_loader_modules(self):
        return {localemod: {}}

    def test_list_avail(self):
        """
        Test for Lists available (compiled) locales
        """
        with patch.dict(
            localemod.__salt__, {"cmd.run": MagicMock(return_value="A\nB")}
        ):
            assert localemod.list_avail() == ["A", "B"]

    @pytest.mark.skipif(_check_localectl, reason="localectl is in degraded state")
    @patch("salt.utils.path.which", MagicMock(return_value="/usr/bin/localctl"))
    @patch(
        "salt.modules.localemod.__salt__",
        {"cmd.run": MagicMock(return_value=locale_ctl_out)},
    )
    def test_localectl_status_parser(self):
        """
        Test localectl status parser.
        :return:
        """
        out = localemod._localectl_status()
        assert isinstance(out, dict)
        for key in ["system_locale", "vc_keymap", "x11_layout", "x11_model"]:
            assert key in out
        assert isinstance(out["system_locale"], dict)
        assert "LANG" in out["system_locale"]
        assert "LANGUAGE" in out["system_locale"]
        assert (
            out["system_locale"]["LANG"]
            == out["system_locale"]["LANGUAGE"]
            == "de_DE.utf8"
        )
        assert isinstance(out["vc_keymap"], dict)
        assert "data" in out["vc_keymap"]
        assert out["vc_keymap"]["data"] is None
        assert isinstance(out["x11_layout"], dict)
        assert "data" in out["x11_layout"]
        assert out["x11_layout"]["data"] == "us"
        assert isinstance(out["x11_model"], dict)
        assert "data" in out["x11_model"]
        assert out["x11_model"]["data"] == "pc105"

    @pytest.mark.skipif(_check_localectl, reason="localectl is in degraded state")
    @patch("salt.utils.path.which", MagicMock(return_value="/usr/bin/localctl"))
    @patch(
        "salt.modules.localemod.__salt__",
        {"cmd.run": MagicMock(return_value=locale_ctl_notset)},
    )
    def test_localectl_status_parser_notset(self):
        """
        Test localectl status parser.
        :return:
        """
        out = localemod._localectl_status()
        assert isinstance(out, dict)
        for key in ["system_locale", "vc_keymap", "x11_layout"]:
            assert key in out
        assert isinstance(out["system_locale"], dict)
        assert "data" in out["system_locale"]
        assert out["system_locale"]["data"] is None
        assert isinstance(out["vc_keymap"], dict)
        assert "data" in out["vc_keymap"]
        assert out["vc_keymap"]["data"] is None
        assert isinstance(out["x11_layout"], dict)
        assert "data" in out["x11_layout"]
        assert out["x11_layout"]["data"] is None

    @patch("salt.modules.localemod.dbus", MagicMock())
    def test_dbus_locale_parser_matches(self):
        """
        Test dbus locale status parser matching the results.
        :return:
        """
        i_dbus = MagicMock()
        i_dbus.Get = MagicMock(return_value=["LANG=de_DE.utf8"])
        dbus = MagicMock(return_value=i_dbus)

        with patch("salt.modules.localemod.dbus.Interface", dbus):
            out = localemod._parse_dbus_locale()
            assert isinstance(out, dict)
            assert "LANG" in out
            assert out["LANG"] == "de_DE.utf8"

    @patch("salt.modules.localemod.dbus", MagicMock())
    @patch("salt.modules.localemod.log", MagicMock())
    def test_dbus_locale_parser_doesnot_matches(self):
        """
        Test dbus locale status parser does not matching the results.
        :return:
        """
        i_dbus = MagicMock()
        i_dbus.Get = MagicMock(return_value=["Fatal error right in front of screen"])
        dbus = MagicMock(return_value=i_dbus)

        with patch("salt.modules.localemod.dbus.Interface", dbus):
            out = localemod._parse_dbus_locale()
            assert isinstance(out, dict)
            assert "LANG" not in out
            assert localemod.log.error.called
            msg = (
                localemod.log.error.call_args[0][0]
                % localemod.log.error.call_args[0][1]
            )
            assert (
                msg == 'Odd locale parameter "Fatal error right in front of screen"'
                " detected in dbus locale output. This should not happen. You should"
                " probably investigate what caused this."
            )

    @patch("salt.utils.path.which", MagicMock(return_value=None))
    @patch("salt.modules.localemod.log", MagicMock())
    def test_localectl_status_parser_no_systemd(self):
        """
        Test localectl status parser raises an exception if no systemd installed.
        :return:
        """
        with pytest.raises(CommandExecutionError) as exc_info:
            localemod._localectl_status()
        assert 'Unable to find "localectl"' in str(exc_info.value)
        assert not localemod.log.debug.called

    @pytest.mark.skipif(_check_localectl, reason="localectl is in degraded state")
    @patch("salt.utils.path.which", MagicMock(return_value="/usr/bin/localctl"))
    @patch(
        "salt.modules.localemod.__salt__",
        {"cmd.run": MagicMock(return_value=locale_ctl_out_empty)},
    )
    def test_localectl_status_parser_empty(self):
        with pytest.raises(CommandExecutionError) as exc_info:
            localemod._localectl_status()
        assert 'Unable to parse result of "localectl"' in str(exc_info.value)

    @pytest.mark.skipif(_check_localectl, reason="localectl is in degraded state")
    @patch("salt.utils.path.which", MagicMock(return_value="/usr/bin/localctl"))
    @patch(
        "salt.modules.localemod.__salt__",
        {"cmd.run": MagicMock(return_value=locale_ctl_out_broken)},
    )
    def test_localectl_status_parser_broken(self):
        with pytest.raises(CommandExecutionError) as exc_info:
            localemod._localectl_status()
        assert 'Unable to parse result of "localectl"' in str(exc_info.value)

    @pytest.mark.skipif(_check_localectl, reason="localectl is in degraded state")
    @patch("salt.utils.path.which", MagicMock(return_value="/usr/bin/localctl"))
    @patch(
        "salt.modules.localemod.__salt__",
        {"cmd.run": MagicMock(return_value=locale_ctl_out_structure)},
    )
    def test_localectl_status_parser_structure(self):
        out = localemod._localectl_status()
        assert isinstance(out, dict)
        for key in ["main", "cow_say"]:
            assert isinstance(out[key], dict)
            for in_key in out[key]:
                assert isinstance(out[key][in_key], str)
        assert isinstance(out["reason"]["data"], str)

    @patch("salt.utils.path.which", MagicMock(return_value="/usr/bin/localctl"))
    @patch(
        "salt.modules.localemod.__grains__",
        {"os_family": "Ubuntu", "osmajorrelease": 42},
    )
    @patch("salt.modules.localemod.dbus", None)
    @patch(
        "salt.modules.localemod._parse_dbus_locale",
        MagicMock(return_value={"LANG": "en_US.utf8"}),
    )
    @patch(
        "salt.modules.localemod._localectl_status",
        MagicMock(return_value={"system_locale": {"LANG": "de_DE.utf8"}}),
    )
    @patch("salt.utils.systemd.booted", MagicMock(return_value=True))
    def test_get_locale_with_systemd_nodbus(self):
        """
        Test getting current system locale with systemd but no dbus available.
        :return:
        """
        assert localemod.get_locale() == "de_DE.utf8"

    @patch("salt.utils.path.which", MagicMock(return_value="/usr/bin/localctl"))
    @patch(
        "salt.modules.localemod.__grains__",
        {"os_family": "Ubuntu", "osmajorrelease": 42},
    )
    @patch("salt.modules.localemod.dbus", True)
    @patch(
        "salt.modules.localemod._parse_dbus_locale",
        MagicMock(return_value={"LANG": "en_US.utf8"}),
    )
    @patch(
        "salt.modules.localemod._localectl_status",
        MagicMock(return_value={"system_locale": {"LANG": "de_DE.utf8"}}),
    )
    @patch("salt.utils.systemd.booted", MagicMock(return_value=True))
    def test_get_locale_with_systemd_and_dbus(self):
        """
        Test getting current system locale with systemd and dbus available.
        :return:
        """
        assert localemod.get_locale() == "en_US.utf8"

    @patch("salt.utils.path.which", MagicMock(return_value="/usr/bin/localctl"))
    @patch(
        "salt.modules.localemod.__grains__", {"os_family": "Suse", "osmajorrelease": 12}
    )
    @patch("salt.modules.localemod.dbus", True)
    @patch(
        "salt.modules.localemod._parse_dbus_locale",
        MagicMock(return_value={"LANG": "en_US.utf8"}),
    )
    @patch(
        "salt.modules.localemod._localectl_status",
        MagicMock(return_value={"system_locale": {"LANG": "de_DE.utf8"}}),
    )
    @patch("salt.modules.localemod.__salt__", {"cmd.run": MagicMock()})
    @patch("salt.utils.systemd.booted", MagicMock(return_value=True))
    def test_get_locale_with_systemd_and_dbus_sle12(self):
        """
        Test getting current system locale with systemd and dbus available on SLE12.
        :return:
        """
        localemod.get_locale()
        assert (
            localemod.__salt__["cmd.run"].call_args[0][0]
            == 'grep "^RC_LANG" /etc/sysconfig/language'
        )

    @patch("salt.utils.path.which", MagicMock(return_value=None))
    @patch(
        "salt.modules.localemod.__grains__",
        {"os_family": "RedHat", "osmajorrelease": 12},
    )
    @patch("salt.modules.localemod.dbus", None)
    @patch("salt.modules.localemod.__salt__", {"cmd.run": MagicMock()})
    @patch("salt.utils.systemd.booted", MagicMock(return_value=False))
    def test_get_locale_with_no_systemd_redhat(self):
        """
        Test getting current system locale with systemd and dbus available on RedHat.
        :return:
        """
        localemod.get_locale()
        assert (
            localemod.__salt__["cmd.run"].call_args[0][0]
            == 'grep "^LANG=" /etc/sysconfig/i18n'
        )

    @patch("salt.utils.path.which", MagicMock(return_value=None))
    @patch(
        "salt.modules.localemod.__grains__",
        {"os_family": "Debian", "osmajorrelease": 12},
    )
    @patch("salt.modules.localemod.dbus", None)
    @patch("salt.modules.localemod.__salt__", {"cmd.run": MagicMock()})
    @patch("salt.utils.systemd.booted", MagicMock(return_value=False))
    def test_get_locale_with_no_systemd_debian(self):
        """
        Test getting current system locale with systemd and dbus available on Debian.
        :return:
        """
        localemod.get_locale()
        assert (
            localemod.__salt__["cmd.run"].call_args[0][0]
            == 'grep "^LANG=" /etc/default/locale'
        )

    @patch("salt.utils.path.which", MagicMock(return_value=None))
    @patch(
        "salt.modules.localemod.__grains__",
        {"os_family": "Gentoo", "osmajorrelease": 12},
    )
    @patch("salt.modules.localemod.dbus", None)
    @patch("salt.modules.localemod.__salt__", {"cmd.run": MagicMock()})
    @patch("salt.utils.systemd.booted", MagicMock(return_value=False))
    def test_get_locale_with_no_systemd_gentoo(self):
        """
        Test getting current system locale with systemd and dbus available on Gentoo.
        :return:
        """
        localemod.get_locale()
        assert (
            localemod.__salt__["cmd.run"].call_args[0][0]
            == "eselect --brief locale show"
        )

    @patch("salt.utils.path.which", MagicMock(return_value=None))
    @patch(
        "salt.modules.localemod.__grains__",
        {"os_family": "Solaris", "osmajorrelease": 12},
    )
    @patch("salt.modules.localemod.dbus", None)
    @patch("salt.modules.localemod.__salt__", {"cmd.run": MagicMock()})
    @patch("salt.utils.systemd.booted", MagicMock(return_value=False))
    def test_get_locale_with_no_systemd_solaris(self):
        """
        Test getting current system locale with systemd and dbus available on Solaris.
        :return:
        """
        localemod.get_locale()
        assert (
            localemod.__salt__["cmd.run"].call_args[0][0]
            == 'grep "^LANG=" /etc/default/init'
        )

    @patch("salt.utils.path.which", MagicMock(return_value=None))
    @patch(
        "salt.modules.localemod.__grains__",
        {"os_family": "BSD", "osmajorrelease": 8, "oscodename": "DrunkDragon"},
    )
    @patch("salt.modules.localemod.dbus", None)
    @patch("salt.modules.localemod.__salt__", {"cmd.run": MagicMock()})
    @patch("salt.utils.systemd.booted", MagicMock(return_value=False))
    def test_get_locale_with_no_systemd_unknown(self):
        """
        Test getting current system locale with systemd and dbus available on Gentoo.
        :return:
        """
        with pytest.raises(CommandExecutionError) as exc_info:
            localemod.get_locale()
        assert '"DrunkDragon" is unsupported' in str(exc_info.value)

    @patch("salt.utils.path.which", MagicMock(return_value="/usr/bin/localctl"))
    @patch(
        "salt.modules.localemod.__grains__",
        {"os_family": "Ubuntu", "osmajorrelease": 42},
    )
    @patch("salt.modules.localemod.dbus", None)
    @patch("salt.utils.systemd.booted", MagicMock(return_value=True))
    @patch("salt.modules.localemod._localectl_set", MagicMock())
    def test_set_locale_with_systemd_nodbus(self):
        """
        Test setting current system locale with systemd but no dbus available.
        :return:
        """
        loc = "de_DE.utf8"
        localemod.set_locale(loc)
        assert localemod._localectl_set.call_args[0][0] == "de_DE.utf8"

    @patch("salt.utils.path.which", MagicMock(return_value="/usr/bin/localctl"))
    @patch(
        "salt.modules.localemod.__grains__",
        {"os_family": "Ubuntu", "osmajorrelease": 42},
    )
    @patch("salt.modules.localemod.dbus", True)
    @patch("salt.utils.systemd.booted", MagicMock(return_value=True))
    @patch("salt.modules.localemod._localectl_set", MagicMock())
    def test_set_locale_with_systemd_and_dbus(self):
        """
        Test setting current system locale with systemd and dbus available.
        :return:
        """
        loc = "de_DE.utf8"
        localemod.set_locale(loc)
        assert localemod._localectl_set.call_args[0][0] == "de_DE.utf8"

    @patch("salt.utils.path.which", MagicMock(return_value="/usr/bin/localctl"))
    @patch(
        "salt.modules.localemod.__grains__", {"os_family": "Suse", "osmajorrelease": 12}
    )
    @patch("salt.modules.localemod.dbus", True)
    @patch("salt.modules.localemod.__salt__", MagicMock())
    @patch("salt.modules.localemod._localectl_set", MagicMock())
    @patch("salt.utils.systemd.booted", MagicMock(return_value=True))
    def test_set_locale_with_systemd_and_dbus_sle12(self):
        """
        Test setting current system locale with systemd and dbus available on SLE12.
        :return:
        """
        loc = "de_DE.utf8"
        localemod.set_locale(loc)
        assert not localemod._localectl_set.called
        assert localemod.__salt__["file.replace"].called
        assert (
            localemod.__salt__["file.replace"].call_args[0][0]
            == "/etc/sysconfig/language"
        )
        assert localemod.__salt__["file.replace"].call_args[0][1] == "^RC_LANG=.*"
        assert localemod.__salt__["file.replace"].call_args[0][2] == f'RC_LANG="{loc}"'

    @patch("salt.utils.path.which", MagicMock(return_value=None))
    @patch(
        "salt.modules.localemod.__grains__",
        {"os_family": "RedHat", "osmajorrelease": 42},
    )
    @patch("salt.modules.localemod.dbus", None)
    @patch("salt.modules.localemod.__salt__", MagicMock())
    @patch("salt.modules.localemod._localectl_set", MagicMock())
    @patch("salt.utils.systemd.booted", MagicMock(return_value=False))
    def test_set_locale_with_no_systemd_redhat(self):
        """
        Test setting current system locale with systemd and dbus available on RedHat.
        :return:
        """
        loc = "de_DE.utf8"
        localemod.set_locale(loc)
        assert not localemod._localectl_set.called
        assert localemod.__salt__["file.replace"].called
        assert (
            localemod.__salt__["file.replace"].call_args[0][0] == "/etc/sysconfig/i18n"
        )
        assert localemod.__salt__["file.replace"].call_args[0][1] == "^LANG=.*"
        assert localemod.__salt__["file.replace"].call_args[0][2] == 'LANG="{}"'.format(
            loc
        )

    @patch("salt.utils.path.which", MagicMock(return_value="/usr/sbin/update-locale"))
    @patch(
        "salt.modules.localemod.__grains__",
        {"os_family": "Debian", "osmajorrelease": 42},
    )
    @patch("salt.modules.localemod.dbus", None)
    @patch("salt.modules.localemod.__salt__", MagicMock())
    @patch("salt.modules.localemod._localectl_set", MagicMock())
    @patch("salt.utils.systemd.booted", MagicMock(return_value=False))
    def test_set_locale_with_no_systemd_debian(self):
        """
        Test setting current system locale with systemd and dbus available on Debian.
        :return:
        """
        loc = "de_DE.utf8"
        localemod.set_locale(loc)
        assert not localemod._localectl_set.called
        assert localemod.__salt__["file.replace"].called
        assert (
            localemod.__salt__["file.replace"].call_args[0][0] == "/etc/default/locale"
        )
        assert localemod.__salt__["file.replace"].call_args[0][1] == "^LANG=.*"
        assert localemod.__salt__["file.replace"].call_args[0][2] == 'LANG="{}"'.format(
            loc
        )

    @patch("salt.utils.path.which", MagicMock(return_value=None))
    @patch("salt.utils.path.which", MagicMock(return_value=None))
    @patch(
        "salt.modules.localemod.__grains__",
        {"os_family": "Debian", "osmajorrelease": 42},
    )
    @patch("salt.modules.localemod.dbus", None)
    @patch("salt.modules.localemod.__salt__", MagicMock())
    @patch("salt.modules.localemod._localectl_set", MagicMock())
    @patch("salt.utils.systemd.booted", MagicMock(return_value=False))
    def test_set_locale_with_no_systemd_debian_no_update_locale(self):
        """
        Test setting current system locale with systemd and dbus available on Debian but update-locale is not installed.
        :return:
        """
        loc = "de_DE.utf8"
        with pytest.raises(CommandExecutionError) as exc_info:
            localemod.set_locale(loc)
        assert not localemod._localectl_set.called
        assert 'Cannot set locale: "update-locale" was not found.' in str(
            exc_info.value
        )

    @patch("salt.utils.path.which", MagicMock(return_value=None))
    @patch(
        "salt.modules.localemod.__grains__",
        {"os_family": "Gentoo", "osmajorrelease": 42},
    )
    @patch("salt.modules.localemod.dbus", None)
    @patch("salt.modules.localemod.__salt__", MagicMock())
    @patch("salt.modules.localemod._localectl_set", MagicMock())
    @patch("salt.utils.systemd.booted", MagicMock(return_value=False))
    def test_set_locale_with_no_systemd_gentoo(self):
        """
        Test setting current system locale with systemd and dbus available on Gentoo.
        :return:
        """
        loc = "de_DE.utf8"
        localemod.set_locale(loc)
        assert not localemod._localectl_set.called
        assert (
            localemod.__salt__["cmd.retcode"].call_args[0][0]
            == "eselect --brief locale set de_DE.utf8"
        )

    @patch("salt.utils.path.which", MagicMock(return_value=None))
    @patch(
        "salt.modules.localemod.__grains__",
        {"os_family": "Solaris", "osmajorrelease": 42},
    )
    @patch("salt.modules.localemod.dbus", None)
    @patch(
        "salt.modules.localemod.__salt__",
        {
            "locale.list_avail": MagicMock(return_value=["de_DE.utf8"]),
            "file.replace": MagicMock(),
        },
    )
    @patch("salt.modules.localemod._localectl_set", MagicMock())
    @patch("salt.utils.systemd.booted", MagicMock(return_value=False))
    def test_set_locale_with_no_systemd_solaris_with_list_avail(self):
        """
        Test setting current system locale with systemd and dbus available on Solaris.
        The list_avail returns the proper locale.
        :return:
        """
        loc = "de_DE.utf8"
        localemod.set_locale(loc)
        assert not localemod._localectl_set.called
        assert localemod.__salt__["file.replace"].called
        assert localemod.__salt__["file.replace"].call_args[0][0] == "/etc/default/init"
        assert localemod.__salt__["file.replace"].call_args[0][1] == "^LANG=.*"
        assert localemod.__salt__["file.replace"].call_args[0][2] == 'LANG="{}"'.format(
            loc
        )

    @patch("salt.utils.path.which", MagicMock(return_value=None))
    @patch(
        "salt.modules.localemod.__grains__",
        {"os_family": "Solaris", "osmajorrelease": 42},
    )
    @patch("salt.modules.localemod.dbus", None)
    @patch(
        "salt.modules.localemod.__salt__",
        {
            "locale.list_avail": MagicMock(return_value=["en_GB.utf8"]),
            "file.replace": MagicMock(),
        },
    )
    @patch("salt.modules.localemod._localectl_set", MagicMock())
    @patch("salt.utils.systemd.booted", MagicMock(return_value=False))
    def test_set_locale_with_no_systemd_solaris_without_list_avail(self):
        """
        Test setting current system locale with systemd and dbus is not available on Solaris.
        The list_avail does not return the proper locale.
        :return:
        """
        loc = "de_DE.utf8"
        assert not localemod.set_locale(loc)
        assert not localemod._localectl_set.called
        assert not localemod.__salt__["file.replace"].called

    @patch("salt.utils.path.which", MagicMock(return_value=None))
    @patch(
        "salt.modules.localemod.__grains__", {"os_family": "BSD", "osmajorrelease": 42}
    )
    @patch("salt.modules.localemod.dbus", None)
    @patch(
        "salt.modules.localemod.__salt__",
        {
            "locale.list_avail": MagicMock(return_value=["en_GB.utf8"]),
            "file.replace": MagicMock(),
        },
    )
    @patch("salt.modules.localemod._localectl_set", MagicMock())
    @patch("salt.utils.systemd.booted", MagicMock(return_value=False))
    def test_set_locale_with_no_systemd_unknown(self):
        """
        Test setting current system locale without systemd on unknown system.
        :return:
        """
        with pytest.raises(CommandExecutionError) as exc_info:
            localemod.set_locale("de_DE.utf8")
        assert "Unsupported platform" in str(exc_info.value)

    @patch(
        "salt.utils.locales.normalize_locale",
        MagicMock(return_value="en_US.UTF-8 UTF-8"),
    )
    @patch(
        "salt.modules.localemod.__salt__",
        {"locale.list_avail": MagicMock(return_value=["A", "B"])},
    )
    def test_avail(self):
        """
        Test for Check if a locale is available
        """
        assert localemod.avail("locale")

    @patch("salt.modules.localemod.log", MagicMock())
    @patch("salt.utils.path.which", MagicMock(return_value="/some/dir/path"))
    @patch("salt.modules.localemod.__grains__", {"os": "Debian"})
    @patch(
        "salt.modules.localemod.__salt__",
        {"file.search": MagicMock(return_value=False)},
    )
    def test_gen_locale_not_valid(self):
        """
        Tests the return of gen_locale when the provided locale is not found
        """
        assert not localemod.gen_locale("foo")
        assert localemod.log.error.called
        msg = localemod.log.error.call_args[0][0] % (
            localemod.log.error.call_args[0][1],
            localemod.log.error.call_args[0][2],
        )
        assert (
            msg == 'The provided locale "foo" is not found in /usr/share/i18n/SUPPORTED'
        )

    @patch("salt.modules.localemod.log", MagicMock())
    @patch("salt.modules.localemod.__grains__", {"os_family": "Suse"})
    @patch("os.listdir", MagicMock(return_value=[]))
    @patch("salt.utils.locales.join_locale", MagicMock(return_value="en_GB.utf8"))
    def test_gen_locale_suse_invalid(self):
        """
        Tests the location where gen_locale is searching for generated paths.
        :return:
        """
        assert not localemod.gen_locale("de_DE.utf8")
        assert localemod.log.error.called
        msg = localemod.log.error.call_args[0][0] % (
            localemod.log.error.call_args[0][1],
            localemod.log.error.call_args[0][2],
        )
        assert localemod.os.listdir.call_args[0][0] == "/usr/share/locale"
        assert (
            msg == 'The provided locale "en_GB.utf8" is not found in /usr/share/locale'
        )

    @patch("salt.modules.localemod.log", MagicMock())
    @patch("salt.modules.localemod.__grains__", {"os_family": "Suse"})
    @patch(
        "salt.modules.localemod.__salt__",
        {"cmd.run_all": MagicMock(return_value={"retcode": 0})},
    )
    @patch("os.listdir", MagicMock(return_value=["de_DE"]))
    @patch("os.path.exists", MagicMock(return_value=False))
    @patch("salt.utils.locales.join_locale", MagicMock(return_value="de_DE.utf8"))
    @patch("salt.utils.path.which", MagicMock(side_effect=[None, "/usr/bin/localedef"]))
    def test_gen_locale_suse_valid(self):
        """
        Tests the location where gen_locale is calling localedef on Suse os-family.
        :return:
        """
        localemod.gen_locale("de_DE.utf8")
        assert localemod.__salt__["cmd.run_all"].call_args[0][0] == [
            "localedef",
            "--force",
            "-i",
            "de_DE",
            "-f",
            "utf8",
            "de_DE.utf8",
            "--quiet",
        ]

    @patch("salt.modules.localemod.log", MagicMock())
    @patch("salt.modules.localemod.__grains__", {"os_family": "Suse"})
    @patch(
        "salt.modules.localemod.__salt__",
        {"cmd.run_all": MagicMock(return_value={"retcode": 0})},
    )
    @patch("os.listdir", MagicMock(return_value=["de_DE"]))
    @patch("os.path.exists", MagicMock(return_value=False))
    @patch("salt.utils.locales.join_locale", MagicMock(return_value="de_DE.utf8"))
    @patch("salt.utils.path.which", MagicMock(return_value=None))
    def test_gen_locale_suse_localedef_error_handling(self):
        """
        Tests the location where gen_locale is handling error while calling not installed localedef on Suse os-family.
        :return:
        """
        with pytest.raises(CommandExecutionError) as exc_info:
            localemod.gen_locale("de_DE.utf8")
        assert (
            'Command "locale-gen" or "localedef" was not found on this system.'
            in str(exc_info.value)
        )

    def test_gen_locale_debian(self):
        """
        Tests the return of successful gen_locale on Debian system
        """
        ret = {"stdout": "saltines", "stderr": "biscuits", "retcode": 0, "pid": 1337}
        with patch.dict(localemod.__grains__, {"os": "Debian"}), patch(
            "salt.utils.path.which", MagicMock(return_value="/some/dir/path")
        ), patch.dict(
            localemod.__salt__,
            {
                "file.search": MagicMock(return_value=True),
                "file.replace": MagicMock(return_value=True),
                "cmd.run_all": MagicMock(return_value=ret),
            },
        ):
            assert localemod.gen_locale("en_US.UTF-8 UTF-8")

    def test_gen_locale_debian_no_charmap(self):
        """
        Tests the return of successful gen_locale on Debian system without a charmap
        """
        ret = {"stdout": "saltines", "stderr": "biscuits", "retcode": 0, "pid": 1337}
        with patch.dict(localemod.__grains__, {"os": "Debian"}), patch(
            "salt.utils.path.which", MagicMock(return_value="/some/dir/path")
        ), patch.dict(
            localemod.__salt__,
            {
                "file.search": lambda s, p, flags: not len(p.split()) == 1,
                "file.replace": MagicMock(return_value=True),
                "cmd.run_all": MagicMock(return_value=ret),
            },
        ):
            assert localemod.gen_locale("en_US.UTF-8")

    def test_gen_locale_ubuntu(self):
        """
        Test the return of successful gen_locale on Ubuntu system
        """
        ret = {"stdout": "saltines", "stderr": "biscuits", "retcode": 0, "pid": 1337}
        with patch.dict(
            localemod.__salt__,
            {
                "file.replace": MagicMock(return_value=True),
                "file.touch": MagicMock(return_value=None),
                "file.append": MagicMock(return_value=None),
                "cmd.run_all": MagicMock(return_value=ret),
            },
        ), patch(
            "salt.utils.path.which", MagicMock(return_value="/some/dir/path")
        ), patch(
            "os.listdir", MagicMock(return_value=["en_US"])
        ), patch.dict(
            localemod.__grains__, {"os": "Ubuntu"}
        ):
            assert localemod.gen_locale("en_US.UTF-8")

    def test_gen_locale_gentoo(self):
        """
        Tests the return of successful gen_locale on Gentoo system
        """
        ret = {"stdout": "saltines", "stderr": "biscuits", "retcode": 0, "pid": 1337}
        with patch.dict(localemod.__grains__, {"os_family": "Gentoo"}), patch(
            "salt.utils.path.which", MagicMock(return_value="/some/dir/path")
        ), patch("os.listdir", MagicMock(return_value=["en_US.UTF-8"])), patch.dict(
            localemod.__salt__,
            {
                "file.search": MagicMock(return_value=True),
                "file.replace": MagicMock(return_value=True),
                "cmd.run_all": MagicMock(return_value=ret),
            },
        ):
            assert localemod.gen_locale("en_US.UTF-8 UTF-8")

    def test_gen_locale_gentoo_no_charmap(self):
        """
        Tests the return of successful gen_locale on Gentoo system without a charmap
        """

        def file_search(search, pattern, flags):
            """
            mock file.search
            """
            if len(pattern.split()) == 1:
                return False
            else:  # charmap was supplied
                return True

        ret = {"stdout": "saltines", "stderr": "biscuits", "retcode": 0, "pid": 1337}
        with patch.dict(localemod.__grains__, {"os_family": "Gentoo"}), patch(
            "salt.utils.path.which", MagicMock(return_value="/some/dir/path")
        ), patch("os.listdir", MagicMock(return_value=["en_US.UTF-8"])), patch.dict(
            localemod.__salt__,
            {
                "file.search": file_search,
                "file.replace": MagicMock(return_value=True),
                "cmd.run_all": MagicMock(return_value=ret),
            },
        ):
            assert localemod.gen_locale("en_US.UTF-8")

    def test_gen_locale(self):
        """
        Tests the return of successful gen_locale
        """
        ret = {"stdout": "saltines", "stderr": "biscuits", "retcode": 0, "pid": 1337}
        with patch.dict(
            localemod.__salt__,
            {"cmd.run_all": MagicMock(return_value=ret), "file.replace": MagicMock()},
        ), patch(
            "salt.utils.path.which", MagicMock(return_value="/some/dir/path")
        ), patch(
            "os.listdir", MagicMock(return_value=["en_US"])
        ):
            assert localemod.gen_locale("en_US.UTF-8")

    def test_gen_locale_verbose(self):
        """
        Tests the return of successful gen_locale
        """
        ret = {"stdout": "saltines", "stderr": "biscuits", "retcode": 0, "pid": 1337}
        with patch.dict(
            localemod.__salt__,
            {"cmd.run_all": MagicMock(return_value=ret), "file.replace": MagicMock()},
        ), patch(
            "salt.utils.path.which", MagicMock(return_value="/some/dir/path")
        ), patch(
            "os.listdir", MagicMock(return_value=["en_US"])
        ):
            assert localemod.gen_locale("en_US.UTF-8", verbose=True) == ret

    @pytest.mark.skipif(_check_localectl, reason="localectl is in degraded state")
    @patch("salt.utils.path.which", MagicMock(return_value="/usr/bin/localctl"))
    def test_parse_localectl(self):
        localectl_out = (
            "   System Locale: LANG=en_US.UTF-8\n"
            "                  LANGUAGE=en_US:en\n"
            "       VC Keymap: n/a"
        )
        mock_cmd = Mock(return_value=localectl_out)
        with patch.dict(localemod.__salt__, {"cmd.run": mock_cmd}):
            ret = localemod._localectl_status()["system_locale"]
            assert {"LANG": "en_US.UTF-8", "LANGUAGE": "en_US:en"} == ret



"""
Tests for file.comment state function
"""

import re

import pytest

import salt.utils.files
from tests.support.helpers import dedent

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.slow_test,
]


@pytest.fixture(scope="module")
def file(states):
    return states.file


@pytest.fixture(scope="function")
def source():
    with pytest.helpers.temp_file(
        name="file.txt",
        contents=dedent(
            """
            things = stuff
            port = 5432                             # (change requires restart)
            # commented = something
            moar = things
            """
        ),
    ) as source:
        yield source


def test_issue_62121(file, source):
    """
    Test file.comment when the comment character is
    later in the line, after the text
    """
    regex = r"^port\s*=.+"
    reg_cmp = re.compile(regex, re.MULTILINE)
    cmt_regex = r"^#port\s*=.+"
    cmt_cmp = re.compile(cmt_regex, re.MULTILINE)

    with salt.utils.files.fopen(str(source)) as _fp:
        assert reg_cmp.findall(_fp.read())

    file.comment(name=str(source), regex=regex)

    with salt.utils.files.fopen(str(source)) as _fp:
        assert not reg_cmp.findall(_fp.read())
        _fp.seek(0)
        assert cmt_cmp.findall(_fp.read())


def test_comment(file, tmp_path):
    """
    file.comment
    """
    name = tmp_path / "testfile"
    # write a line to file
    name.write_text("comment_me")

    # Look for changes with test=True: return should be "None" at the first run
    ret = file.comment(name=str(name), regex="^comment", test=True)
    assert ret.result is None

    # comment once
    ret = file.comment(name=str(name), regex="^comment")
    assert ret.result is True
    # line is commented
    assert name.read_text().startswith("#comment")

    # comment twice
    ret = file.comment(name=str(name), regex="^comment")
    # result is still positive
    assert ret.result is True
    # line is still commented
    assert name.read_text().startswith("#comment")

    # Test previously commented file returns "True" now and not "None" with test=True
    ret = file.comment(name=str(name), regex="^comment", test=True)
    assert ret.result is True


def test_test_comment(file, tmp_path):
    """
    file.comment test interface
    """
    name = tmp_path / "testfile"
    # write a line to file
    name.write_text("comment_me")

    # Look for changes with test=True: return should be "None" at the first run
    ret = file.comment(name=str(name), regex="^comment", test=True)
    assert ret.result is None
    assert "#comment" not in name.read_text()


def test_issue_2401_file_comment(modules, tmp_path):
    # Get a path to the temporary file
    tmp_file = tmp_path / "issue-2041-comment.txt"
    tmp_file.write_text("hello\nworld\n")
    # create the sls template
    template_lines = [
        f"{tmp_file}:",
        "  file.comment:",
        "    - regex: ^world",
    ]
    template = "\n".join(template_lines)
    ret = modules.state.template_str(template)
    for state_run in ret:
        assert state_run.result is True
        assert "Pattern already commented" not in state_run.comment
        assert "Commented lines successfully" in state_run.comment

    # This next time, it is already commented.
    ret = modules.state.template_str(template)
    for state_run in ret:
        assert state_run.result is True
        assert "Pattern already commented" in state_run.comment


def test_issue_65501(file, tmp_path):
    tmp_file = tmp_path / "issue-65501.txt"
    tmp_file.write_text("first\n#PermitRootLogin prohibit-password\nlast")
    ret = file.comment(
        name=str(tmp_file),
        regex="^PermitRootLogin[ \t]+.*$",
        char="# NEXT LINE COMMENT SALTSTACK openssh-server_comment_permitrootlogin_sshd_config\n# ",
        ignore_missing=True,
    )
    assert ret.result is True
    assert ret.comment == "Pattern not found and ignore_missing set to True"



import hashlib
import hmac
import os

import pytest
from cryptography.hazmat.backends.openssl import backend
from cryptography.hazmat.primitives import serialization

import salt.crypt as crypt
import salt.utils.files
import salt.utils.stringutils
from tests.conftest import FIPS_TESTRUN
from tests.support.mock import mock_open, patch

from . import (
    HAS_M2,
    MSG,
    PRIVKEY_DATA,
    PUBKEY_DATA,
    SIG,
    LegacyPrivateKey,
    LegacyPublicKey,
    legacy_gen_keys,
)

if HAS_M2:
    from . import EVP, RSA
else:
    from . import AES, PKCS1_OAEP


OPENSSL_ENCRYPTED_KEY = """
-----BEGIN RSA PRIVATE KEY-----
Proc-Type: 4,ENCRYPTED
DEK-Info: AES-256-CBC,12E2FE8CA93672B629477073867DE7E4

3wgKQlgjT3G+8rIQt97AXMDByh5u9JMZPYOB9/wg3iC3qoJXfoFAsCNUjODJBnkI
j9Zgj/bOCaSM3UshMQXmYY+2Rfi1SVQPnETlqEH/plMZS38tB8mN5pBdthgGTw6c
rhpj/S23eZT5d+z5ODeVYlWCVhx8CtE8OQEzOQk8dxLWbVHhvgC3uJGWOPR3P7VM
BlxH5LxWRCrC8vzbnwwaJnX8BTQ7fc4qeGwHlXBjpnxQhxO27pEj08NQ0/lfKh1b
seX5uiCjuQhHFKNGTuA16rQIe6BkYRHIWCDhySftl/lqSfLQif0OAXaHEdL2EdIS
ySD12RYyNUDotEzYFF+qzJ5OAtraqvc8kYror7oN52bHKCjJyrp4+DWA4/N7FjTV
+FqUyJNKqw1DAQAxlZlq+GgNyi8+g/Zs2TKTc/ZXaPLjtWYOQEQkYaNBoaD3ydY3
c7x+uQtJLVW9BF9FSl9A7BItpqZQWKiHiGtUdhYOkemlR+zMatjBe/eTq8LrnEDa
IyI+rRo1PSDAz3n1pdEAzGAOeqwT+j7YG9O8+dybMY5FcAtiiPX21nIpmP+Rtx4X
GqzHsT7nM6QG4O8GPKuK6TniG+Q0doNWomwuau/cjQgL4C+yFiX3kIPpHz9kA/aS
NL1SJqSsvc3D/KlRbHXaJZJyhmzDuEbQynkaAdvejiajlJWAwA3BZWw1RUK7Wn8m
XcNPJL3g02uKq8SUDgVQl/cx4QawuWri2Xh8/xakNYWzNU2feoWBmV+gN2qDSxyz
Qi+xu3CzdJVrPs71lW0rEAIQvU3K3Umava9M4CUF6R7N9+Zv+m1EuMQs0IGt8VCY
Wo1KY5PAb/V718d1C3I6kXvLSDXG8xqyEleilPLhKCRGPK+2g0nGYu562EV1i6by
gr+PLnFJTfgHEzwIfsqfNoR8ReQ6AJKJoniQr4vqex9xtifuhes0odpqmUB4/B2C
UfY/SpJR6tzdrGndpB/vb1vjHumHklHHWrLONtz70BhR8Zaisc7SCmL5bFgWqzMC
MJKPulRRGQCPAzy5OI/ZULY8+dzlva1MyoCYlWjeUtcUAy+9dyA8GZv75ez9g71b
10nNINDcvGG7zWShSYrAKrvLlsoE7eZ+flG+XhI2CfiC9/zHBzy/slbaH9H+1tlO
VWKiw6iBb2TEvBk4Wpk2nUFlWKtkkBVAlgbShbE2K8pTHrJeIRv5J897k693NFZE
DjVVJirzMv/OiZTami0qBQ4nDtUZpH8FsFZ8DtREkhROsDmrjq9PGkOVaxEyF/ke
avJT34gp4OoNWj7/Rd1YNbGiWjMEB3zi9y1Q6Ufiod9ZlK3RQb4tNrpzDn/msdJo
pIkuByWjXjF4YQRKtAoeCn+CWiY7L/Qi8X7jmX27JLILlZPtTJ+aNp3eCr6ZX0dW
y0uhN89sgMewlvDA/pduL/VJRHUBZC2eD8FbD7p6K+yRKhdciS9A8F6aIhx615s6
mngRBfwzh8ST6yQgLwCgle/XaRYTWJKjzAe3lkaIBBhHpeuq1UMAjunoS8JnLNiy
xQJ0PznzY57sYKpIiClwMjfpnX47nTU2DFWuPEXvBtG1xMjacGPbVrUslesY5bii
-----END RSA PRIVATE KEY-----
"""


@pytest.fixture
def openssl_encrypted_key():
    return OPENSSL_ENCRYPTED_KEY


@pytest.fixture
def passphrase():
    return "pass1234"


@pytest.fixture
def signing_algorithm():
    if FIPS_TESTRUN:
        return salt.crypt.PKCS1v15_SHA224
    return salt.crypt.PKCS1v15_SHA1


@pytest.fixture
def encryption_algorithm():
    if FIPS_TESTRUN:
        return salt.crypt.OAEP_SHA224
    return salt.crypt.OAEP_SHA1


SIG_SHA224 = (
    b'\x18)\xc3E|\x15\xebF\x0f\xe6\xc0\x10\xca\xd9~\x1d\xf14t\xc7\x14}\xda6Fk"#'
    b'Hl\x06\x13\xa9\xe3QlL\\\xf4`r\x88\x85\xc6#s\xcb"6\x1c\xdd\x07t\xd4\x84g'
    b"n\x0f\xcc\x1c\xee\xe7\x84T\xb7\xd1\xc80~\xdd\xf7+\x972b6\xf1\xe1\x00P"
    b"E\xb8\x86\xb3i\xa6*\xd2\xac\xb5\xcbStg\xfb*E9+\xf7\xc5\xc6X\x1e\xb9vY\xb7"
    b"kT[a\xe8\xe1\xd8\xdf'u\x00k\x13\xff\xe2\xd1\x91M\xa7U\xc9\x90z\xf0"
    b"\x03\xb2\xf3\x1bR\xbd\xc8\xe4B\xadJ\x91\x1e\x98\xea\x17\xa8;\x01\xcb"
    b"1\x07\x7f\xa2\xf3\xe6\x83\xed\x03m\xad\t&\x95\xc2Q\xfcs\xcbV\xd4\xa4\xc9n"
    b"\x8a\xbe\xcc3?.N\x1f8d{B\x8cp\xf8\xc8\x17\x90\x0e\x0c\x1a\x8dF\xb8"
    b'\x18\xf7\x97\xf0\x04L\xe6\xfb\xc1\xb0}\xa9\xb6?\xc0\xbd\x8a<\xac"5\xee@x'
    b"\xea\x1d\xa3\xffB\xa5\xbdt`\xa5\xe8p\xa3/\x18+\xec5\xb3]\x92\xaa\xd7\x9c"
    b"\x0b\x03`~\x00\r%\xc8"
)


@pytest.fixture
def signature():
    if FIPS_TESTRUN:
        return SIG_SHA224
    return SIG


@pytest.fixture
def private_key(passphrase, tmp_path):
    keypath = tmp_path / "keys"
    keypath.mkdir()
    keyname = "test"
    keysize = 2048
    return crypt.gen_keys(str(keypath), keyname, keysize, passphrase=passphrase)


def test_fips_mode():
    assert backend._fips_enabled == FIPS_TESTRUN


@pytest.mark.skipif(FIPS_TESTRUN, reason="Legacy key can not be loaded in FIPS mode")
def test_gen_keys_legacy(tmp_path):
    keypath = tmp_path / "keys"
    keypath.mkdir()
    passphrase = "pass1234"
    keyname = "test"
    keysize = 2048
    ret = legacy_gen_keys(str(keypath), keyname, keysize, passphrase=passphrase)
    with salt.utils.files.fopen(ret, "rb") as fp:
        keybytes = fp.read()
        assert keybytes.startswith(b"-----BEGIN RSA PRIVATE KEY-----\n")
        priv = serialization.load_pem_private_key(keybytes, passphrase.encode())
    with salt.utils.files.fopen(ret.replace(".pem", ".pub"), "rb") as fp:
        keybytes = fp.read()
        assert keybytes.startswith(b"-----BEGIN PUBLIC KEY-----\n")


def test_gen_keys(tmp_path):
    keypath = tmp_path / "keys"
    keypath.mkdir()
    passphrase = "pass1234"
    keyname = "test"
    keysize = 2048
    ret = crypt.gen_keys(str(keypath), keyname, keysize, passphrase=passphrase)
    with salt.utils.files.fopen(ret, "rb") as fp:
        keybytes = fp.read()
        if FIPS_TESTRUN:
            assert keybytes.startswith(b"-----BEGIN ENCRYPTED PRIVATE KEY-----\n")
        else:
            assert keybytes.startswith(b"-----BEGIN RSA PRIVATE KEY-----\n")
        priv = serialization.load_pem_private_key(keybytes, passphrase.encode())
    with salt.utils.files.fopen(ret.replace(".pem", ".pub"), "rb") as fp:
        keybytes = fp.read()
        assert keybytes.startswith(b"-----BEGIN PUBLIC KEY-----\n")


def test_legacy_private_key_loading(private_key, passphrase):
    priv = LegacyPrivateKey(private_key.encode(), passphrase.encode())
    assert priv.key


def test_private_key_loading(private_key, passphrase):
    priv = crypt.PrivateKey(private_key, passphrase)
    assert priv.key


@pytest.mark.skipif(FIPS_TESTRUN, reason="Legacy key can not be loaded in FIPS mode")
def test_private_key_signing(private_key, passphrase):
    lpriv = LegacyPrivateKey(private_key.encode(), passphrase.encode())
    priv = crypt.PrivateKey(private_key, passphrase)
    data = b"meh"
    signature = priv.sign(data)
    lsignature = lpriv.sign(data)
    assert lsignature == signature


@pytest.mark.skipif(FIPS_TESTRUN, reason="Legacy key can not be loaded in FIPS mode")
def test_legacy_public_key_verify(private_key, passphrase):
    lpriv = crypt.PrivateKey(private_key, passphrase)
    data = b"meh"
    signature = lpriv.sign(data)
    pubkey = LegacyPublicKey(private_key.replace(".pem", ".pub"))
    assert pubkey.verify(data, signature)


@pytest.mark.skipif(FIPS_TESTRUN, reason="Legacy key can not be loaded in FIPS mode")
def test_public_key_verify(private_key, passphrase):
    lpriv = LegacyPrivateKey(private_key.encode(), passphrase.encode())
    data = b"meh"
    signature = lpriv.sign(data)
    pubkey = crypt.PublicKey(private_key.replace(".pem", ".pub"))
    assert pubkey.verify(data, signature)


@pytest.mark.skipif(FIPS_TESTRUN, reason="Legacy key can not be loaded in FIPS mode")
def test_public_key_encrypt(private_key, passphrase):
    pubkey = crypt.PublicKey(private_key.replace(".pem", ".pub"))
    data = b"meh"
    enc = pubkey.encrypt(data)

    lpriv = LegacyPrivateKey(private_key.encode(), passphrase.encode())
    if HAS_M2:
        dec = lpriv.key.private_decrypt(enc, RSA.pkcs1_oaep_padding)
    else:
        cipher = PKCS1_OAEP.new(lpriv.key)
        dec = cipher.decrypt(enc)

    assert data == dec


@pytest.mark.skipif(FIPS_TESTRUN, reason="Legacy key can not be loaded in FIPS mode")
def test_private_key_decrypt(private_key, passphrase):
    lpubkey = LegacyPublicKey(private_key.replace(".pem", ".pub"))
    data = b"meh"
    enc = lpubkey.encrypt(data)
    priv = crypt.PrivateKey(private_key, passphrase)
    dec = priv.key.decrypt(
        enc,
        crypt.padding.OAEP(
            mgf=crypt.padding.MGF1(algorithm=crypt.hashes.SHA1()),
            algorithm=crypt.hashes.SHA1(),
            label=None,
        ),
    )

    assert data == dec


def test_legacy_aes_encrypt():
    """
    Test that the legacy aes encryption can be decrypted by cryptography
    """
    orig_data = b"meh"
    crypticle = salt.crypt.Crypticle({}, salt.crypt.Crypticle.generate_key_string())
    aes_key, hmac_key = crypticle.keys
    pad = crypticle.AES_BLOCK_SIZE - len(orig_data) % crypticle.AES_BLOCK_SIZE
    data = orig_data + salt.utils.stringutils.to_bytes(pad * chr(pad))
    iv_bytes = os.urandom(crypticle.AES_BLOCK_SIZE)
    iv_bytes = data[: crypticle.AES_BLOCK_SIZE]
    if HAS_M2:
        cypher = EVP.Cipher(
            alg="aes_192_cbc", key=aes_key, iv=iv_bytes, op=1, padding=False
        )
        encr = cypher.update(data)
        encr += cypher.final()
    else:
        cypher = AES.new(aes_key, AES.MODE_CBC, iv_bytes)
        encr = cypher.encrypt(data)

    data = iv_bytes + encr
    sig = hmac.new(hmac_key, data, hashlib.sha256).digest()
    assert orig_data == crypticle.decrypt(data + sig)


def test_aes_encrypt():
    """
    Test that cryptography aes encryption can be decrypted by the legacy libraries
    """
    orig_data = b"meh"
    crypticle = salt.crypt.Crypticle({}, salt.crypt.Crypticle.generate_key_string())

    data = crypticle.encrypt(orig_data)
    aes_key, hmac_key = crypticle.keys
    sig = data[-crypticle.SIG_SIZE :]
    data = data[: -crypticle.SIG_SIZE]
    if not isinstance(data, bytes):
        data = salt.utils.stringutils.to_bytes(data)
    mac_bytes = hmac.new(hmac_key, data, hashlib.sha256).digest()
    result = 0
    for zipped_x, zipped_y in zip(mac_bytes, sig):
        result |= zipped_x ^ zipped_y
    iv_bytes = data[: crypticle.AES_BLOCK_SIZE]
    data = data[crypticle.AES_BLOCK_SIZE :]
    if HAS_M2:
        cypher = EVP.Cipher(
            alg="aes_192_cbc", key=aes_key, iv=iv_bytes, op=0, padding=False
        )
        encr = cypher.update(data)
        data = encr + cypher.final()
    else:
        cypher = AES.new(aes_key, AES.MODE_CBC, iv_bytes)
        data = cypher.decrypt(data)
    data = data[: -data[-1]]
    assert orig_data == data


def test_encrypt_decrypt(private_key, passphrase, encryption_algorithm):
    pubkey = crypt.PublicKey(private_key.replace(".pem", ".pub"))
    enc = pubkey.encrypt(b"meh", algorithm=encryption_algorithm)
    privkey = crypt.PrivateKey(private_key, passphrase)
    assert privkey.decrypt(enc, algorithm=encryption_algorithm) == b"meh"


def test_sign_message(signature, signing_algorithm):
    key = salt.crypt.serialization.load_pem_private_key(PRIVKEY_DATA.encode(), None)
    with patch("salt.crypt.get_rsa_key", return_value=key):
        assert (
            salt.crypt.sign_message(
                "/keydir/keyname.pem", MSG, algorithm=signing_algorithm
            )
            == signature
        )


def test_sign_message_with_passphrase(signature, signing_algorithm):
    key = salt.crypt.serialization.load_pem_private_key(PRIVKEY_DATA.encode(), None)
    with patch("salt.crypt.get_rsa_key", return_value=key):
        assert (
            salt.crypt.sign_message(
                "/keydir/keyname.pem",
                MSG,
                passphrase="password",
                algorithm=signing_algorithm,
            )
            == signature
        )


def test_verify_signature():
    with patch("salt.utils.files.fopen", mock_open(read_data=PUBKEY_DATA.encode())):
        assert salt.crypt.verify_signature("/keydir/keyname.pub", MSG, SIG)


def test_loading_encrypted_openssl_format(openssl_encrypted_key, passphrase, tmp_path):
    path = tmp_path / "key"
    path.write_text(openssl_encrypted_key)
    if FIPS_TESTRUN:
        with pytest.raises(ValueError):
            salt.crypt.get_rsa_key(path, passphrase)
    else:
        try:
            salt.crypt.get_rsa_key(path, passphrase)
        # BaseException to catch errors bubbling up from the cryptogrphy's
        # rust layer.
        except BaseException as exc:  # pylint: disable=broad-except
            pytest.fail(f"Unexpected exception: {exc}")


@pytest.mark.skipif(not FIPS_TESTRUN, reason="Only valid when in FIPS mode")
def test_fips_bad_signing_algo(private_key, passphrase):
    key = salt.crypt.PrivateKey(private_key, passphrase)
    with pytest.raises(salt.exceptions.UnsupportedAlgorithm):
        key.sign("meh", salt.crypt.PKCS1v15_SHA1)


@pytest.mark.skipif(not FIPS_TESTRUN, reason="Only valid when in FIPS mode")
def test_fips_bad_signing_algo_verification(private_key, passphrase):
    lpriv = LegacyPrivateKey(private_key.encode(), passphrase.encode())
    data = b"meh"
    signature = lpriv.sign(data)
    pubkey = salt.crypt.PublicKey(private_key.replace(".pem", ".pub"))
    # cryptogrpahy silently returns False on unsuppoted algorithm
    assert pubkey.verify(signature, salt.crypt.PKCS1v15_SHA1) is False


@pytest.mark.skipif(not FIPS_TESTRUN, reason="Only valid when in FIPS mode")
def test_fips_bad_encryption_algo(private_key, passphrase):
    key = salt.crypt.PublicKey(private_key.replace(".pem", ".pub"))
    with pytest.raises(salt.exceptions.UnsupportedAlgorithm):
        key.encrypt("meh", salt.crypt.OAEP_SHA1)


@pytest.mark.skipif(not FIPS_TESTRUN, reason="Only valid when in FIPS mode")
def test_fips_bad_decryption_algo(private_key, passphrase):
    pubkey = LegacyPublicKey(private_key.replace(".pem", ".pub"))
    data = pubkey.encrypt("meh")
    key = salt.crypt.PrivateKey(private_key, passphrase)
    with pytest.raises(salt.exceptions.UnsupportedAlgorithm):
        key.decrypt(data)



"""
unit tests for salt.utils.job
"""

import salt.minion
import salt.utils.job as job
from tests.support.mock import patch
from tests.support.unit import TestCase


class MockMasterMinion:
    def return_mock_jobs(self):
        return self.mock_jobs_cache

    opts = {"job_cache": True, "ext_job_cache": None, "master_job_cache": "foo"}
    mock_jobs_cache = {}
    returners = {
        "foo.save_load": lambda *args, **kwargs: True,
        "foo.prep_jid": lambda *args, **kwargs: True,
        "foo.get_load": lambda *args, **kwargs: True,
        "foo.returner": lambda *args, **kwargs: True,
    }

    def __init__(self, *args, **kwargs):
        pass


class JobTest(TestCase):
    """
    Validate salt.utils.job
    """

    def test_store_job_exception_handled(self):
        """
        test store_job exception handling
        """
        for func in ["foo.save_load", "foo.prep_jid", "foo.returner"]:

            def raise_exception(*arg, **kwarg):
                raise Exception("expected")

            with patch.object(
                salt.minion, "MasterMinion", MockMasterMinion
            ), patch.dict(MockMasterMinion.returners, {func: raise_exception}), patch(
                "salt.utils.verify.valid_id", return_value=True
            ):
                with self.assertLogs("salt.utils.job", level="CRITICAL") as logged:
                    job.store_job(
                        MockMasterMinion.opts,
                        {
                            "jid": "20190618090114890985",
                            "return": {"success": True},
                            "id": "a",
                        },
                    )
                    self.assertIn(
                        "The specified 'foo' returner threw a stack trace",
                        logged.output[0],
                    )
